"""
ConvMixer.
"""


import typing as T

from flax import linen as nn

from .. import layers
from .factory import register_configs


class ConvActBN(nn.Module):
	"""
	Convolution followed by an activation function and batch normalization.

	Args:
		out_dim (T.Optional[int]): Number of output channels.
		If None, it is set to the number of input channels.
		Default is None.
		kernel_size (T.Union[T.Tuple[int, int], int]): Kernel size.
		If an int, this value is used along both spatial dimensions.
		stride (T.Union[T.Tuple[int, int], int]): Stride. If an int,
		this value is used along both spatial dimensions.
		Default is 1.
		padding (T.Optional[T.Union[str, int]]): Padding. If None,
		it is set so the spatial dimensions are exactly divided by stride.
		If an int, this value is used along both spatial dimensions.
		Default is None.
		groups (T.Union[int, str]): Number of groups. If 'dw', a depthwise convolution is performed.
		Default is 1.
		act (T.Callable): Activation function.
		Default is identity.
	"""
	out_dim: T.Optional[int] = None
	kernel_size: T.Union[T.Tuple[int, int], int] = 3
	stride: T.Union[T.Tuple[int, int], int] = 1
	padding: T.Optional[T.Union[str, int]] = None
	groups: T.Union[int, str] = 1
	act: T.Callable = layers.identity

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = layers.Conv(
			out_dim=self.out_dim,
			kernel_size=self.kernel_size,
			stride=self.stride,
			padding=self.padding,
			groups=self.groups,
			)(input)
		output = self.act(output)
		output = nn.BatchNorm(
			use_running_average=not training,
			)(output)
		return output


class ConvMixerBlock(nn.Module):
	"""
	ConvMixer block.

	Args:
		kernel_size (int): Kernel size.
		Default is 9.
		act (T.Callable): Activation function.
		Default is layers.gelu.
	"""
	kernel_size: int = 9
	act: T.Callable = layers.gelu

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = input + ConvActBN(
				kernel_size=self.kernel_size,
				padding='same',
				groups='dw',
				act=self.act,
				)(input, training=training)
		output = ConvActBN(
				kernel_size=1,
				act=self.act,
				)(output, training=training)
		return output


class ConvMixer(nn.Module):
	"""
	ConvMixer.

	Args:
		depth (int): Depth.
		dim (int): Dimension throughout the network.
		patch_size (int): Patch size.
		Default is 7.
		kernel_size (int): Kernel size.
		Default is 9.
		act (T.Callable): Activation function.
		Default is layers.gelu.
		n_classes (int): Number of output classes. If 0, there is no head,
		and the raw final features are returned. If -1, all stages of the 
		head, other than the final linear layer, are applied and the output
		returned.
		Default is 0.
	"""
	depth: int
	dim: int
	patch_size: int = 7
	kernel_size: int = 9
	act: T.Callable = layers.gelu
	n_classes: int = 0

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = ConvActBN(
			out_dim=self.dim,
			kernel_size=self.patch_size,
			stride=self.patch_size,
			padding=0,
			act=self.act,
			)(input, training=training)
		self.sow(
			col='intermediates',
			name='block_0',
			value=output,
			)
		
		for block_ind in range(self.depth):
			output = ConvMixerBlock(
				kernel_size=self.kernel_size,
				act=self.act,
				)(output, training=training)
			self.sow(
				col='intermediates',
				name=f'block_{block_ind+1}',
				value=output,
				)

		output = layers.Head(
			n_classes=self.n_classes,
			)(output)
		
		return output


@register_configs
def get_convmixer_configs() -> T.Tuple[T.Type[ConvMixer], T.Dict]:
	"""
	Gets configurations for all available
	ConvMixer models.

	Returns (T.Tuple[T.Type[ConvMixer], T.Dict]): The ConvMixer class and
	configurations of all available models.
	"""
	configs = {
		'convmixer20_1024d_patch14_kernel9': {
			'depth': 20,
			'dim': 1024,
			'patch_size': 14,
			},
		'convmixer20_1536d_patch7_kernel9': {
			'depth': 20,
			'dim': 1536,
			},
		'convmixer32_768d_patch7_kernel7': {
			'depth': 32,
			'dim': 768,
			'kernel_size': 7,
			'act': nn.relu,
			},
		}
	return ConvMixer, configs



"""
LayerScale by Touvron et al.

References:
- Touvron et al. Going deeper with Image Transformers.
"""


__all__ = [
	'LayerScale',
	]


import typing as T

from flax import linen as nn
from jax import numpy as jnp


class LayerScale(nn.Module):
	"""
	LayerScale.

	Args:
		init_value (T.Optional[float]): Initial value for the
		lambda parameter. If None, no LayerScale is applied.
		Default is 1e-6.
	"""
	init_value: T.Optional[float] = 1e-6

	@nn.compact
	def __call__(self, input):
		return input * self.param(
			name='lambda',
			init_fn=lambda prng: self.init_value*jnp.ones(input.shape[-1]),
			) if self.init_value else input



"""
Squeeze-and-excitation (SE) by Hu et al.

References:
- Hu et al. Squeeze-and-Excitation Networks.
"""


__all__ = [
	'SE',
	]


import typing as T

from flax import linen as nn

from .mlp import MLP
from .pool import global_avg_pool


class SE(nn.Module):
	"""
	Squeeze-and-excitation.

	Args:
		reduction_factor (T.Optional[int]): Factor of reduction for the 
		hidden layer of the excitation module. If None, reduction_dim
		must be provided. If both reduction_factor and reduction_dim are provided,
		the latter takes precedence.
		Default is 16.
		reduction_dim (T.Optional[int]): Number of channels in the 
		hidden layer of the excitation module. If None, reduction_factor
		must be provided. If both reduction_factor and reduction_dim are provided,
		the latter takes precedence.
		Default is None.
		pool (T.Callable): Pooling method used for aggregating
		the activations of each channel.
		Default is global_avg_pool.
		act (T.Callable): Activation for the excitation module.
		Default is nn.relu.
		gate (T.Callable): Gating function used to normalize the attention
		scores.
		Default is nn.sigmoid.
		bias (bool): Whether the linear layers should have a bias term.
		Default is True.
		bottleneck (bool): If True, SE's usual MLP is used, i.e., one with 
		a single hidden layer. Otherwise, a single fully-connected layer 
		with no non-linearities is used, and reduction_factor, reduction_dim,
		and act are ignored.
		Default is True.
	"""
	reduction_factor: T.Optional[int] = 16
	reduction_dim: T.Optional[int] = None
	pool: T.Callable = global_avg_pool
	act: T.Callable = nn.relu
	gate: T.Callable = nn.sigmoid
	bias: bool = True
	bottleneck: bool = True

	@nn.compact
	def __call__(self, input):
		attention = self.pool(input)

		if self.bottleneck:
			attention = MLP(
				hidden_dim_expansion_factor=1/self.reduction_factor if self.reduction_factor else None,
				hidden_dim=self.reduction_dim,
				act=self.act,
				bias=self.bias,
				)(attention)
		
		else:
			attention = nn.Dense(
				features=input.shape[-1],
				use_bias=self.bias,
				)(attention)
		
		attention = self.gate(attention)
		return attention*input



"""
Tuplifies the input.
"""


__all__ = [
	'tuplify',
	]


import typing as T


def tuplify(
	item: T.Any,
	seq_len: int = 2,
	) -> T.Tuple:
	"""
	Tuplifies the input.

	Args:
		item (T.Any): Item to tuplify. If it
		is a T.Sequence, it is returned unchanged.
		Otherwise, a tuple consisting of seq_len repitions
		of item is returned.
		seq_len (int): How many times to repeat item
		when converting it to a tuple.
	
	Returns (T.Tuple): Tuplified version of item.
	"""
	return item if isinstance(item, T.Sequence) else seq_len*(item,)



"""
Modules following a linear-normalization-activation structure:
- ConvBNAct: Convolution followed by batch normalization and an activation function.
- ConvLNAct: Convolution followed by layer normalization and an activation function.
"""


__all__ = [
	'ConvBNAct',
	'ConvLNAct',
	]


import typing as T

from flax import linen as nn

from .conv import Conv
from .identity import identity


class ConvBNAct(nn.Module):
	"""
	Convolution followed by batch normalization and an activation function.

	Args:
		out_dim (T.Optional[int]): Number of output channels.
		If None, it is set to the number of input channels.
		Default is None.
		kernel_size (T.Union[T.Tuple[int, int], int]): Kernel size.
		If an int, this value is used along both spatial dimensions.
		stride (T.Union[T.Tuple[int, int], int]): Stride. If an int,
		this value is used along both spatial dimensions.
		Default is 1.
		padding (T.Optional[T.Union[str, int]]): Padding. If None,
		it is set so the spatial dimensions are exactly divided by stride.
		If an int, this value is used along both spatial dimensions.
		Default is None.
		groups (T.Union[int, str]): Number of groups. If 'dw', a depthwise convolution is performed.
		Default is 1.
		dilation (int): Dilation.
		Default is 1.
		bias (bool): Whether the convolution should have a bias term if bn is False.
		Default is True.
		bias_force (bool): Whether to force the convolution to have a bias term 
		even if bn is True.
		Default is False.
		bn (bool): Whether to have batch normalization after the convolution.
		If False, the training argument is ignored.
		Default is True.
		act (T.Callable): Activation function.
		Default is identity.
		tf (bool): Whether padding and batch_norm_eps should be fixed at 'same'
		and 1e-3 to conform to TensorFlow's settings.
		Default is False.
	"""
	out_dim: T.Optional[int] = None
	kernel_size: T.Union[T.Tuple[int, int], int] = 3
	stride: T.Union[T.Tuple[int, int], int] = 1
	padding: T.Optional[T.Union[str, int]] = None
	groups: T.Union[int, str] = 1
	dilation: int = 1
	bias: bool = True
	bias_force: bool = False
	bn: bool = True
	act: T.Callable = identity
	tf: bool = False

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = Conv(
			out_dim=self.out_dim,
			kernel_size=self.kernel_size,
			stride=self.stride,
			padding='same' if self.tf else self.padding,
			groups=self.groups,
			dilation=self.dilation,
			bias=self.bias_force or (self.bias and not self.bn),
			)(input)
		output = nn.BatchNorm(
			use_running_average=not training,
			epsilon=1e-3 if self.tf else 1e-5,
			)(output) if self.bn else output
		output = self.act(output)
		return output


class ConvLNAct(nn.Module):
	"""
	Convolution followed by layer normalization and an activation function.

	Args:
		out_dim (T.Optional[int]): Number of output channels.
		If None, it is set to the number of input channels.
		Default is None.
		kernel_size (T.Union[T.Tuple[int, int], int]): Kernel size.
		If an int, this value is used along both spatial dimensions.
		stride (T.Union[T.Tuple[int, int], int]): Stride. If an int,
		this value is used along both spatial dimensions.
		Default is 1.
		padding (T.Optional[T.Union[str, int]]): Padding. If None,
		it is set so the spatial dimensions are exactly divided by stride.
		If an int, this value is used along both spatial dimensions.
		Default is None.
		groups (T.Union[int, str]): Number of groups. If 'dw', a depthwise convolution is performed.
		Default is 1.
		dilation (int): Dilation.
		Default is 1.
		bias (bool): Whether to the convolution should have a bias term.
		Default is True.
		layer_norm_eps (T.Optional[float]): Epsilon value passed to layer normalization.
		If None, no normalization is applied.
		Default is 1e-6.
		act (T.Callable): Activation function.
		Default is identity.
	"""
	out_dim: T.Optional[int] = None
	kernel_size: T.Union[T.Tuple[int, int], int] = 3
	stride: T.Union[T.Tuple[int, int], int] = 1
	padding: T.Optional[T.Union[str, int]] = None
	groups: T.Union[int, str] = 1
	dilation: int = 1
	bias: bool = True
	layer_norm_eps: T.Optional[float] = 1e-6
	act: T.Callable = identity

	@nn.compact
	def __call__(self, input):
		output = Conv(
			out_dim=self.out_dim,
			kernel_size=self.kernel_size,
			stride=self.stride,
			padding=self.padding,
			groups=self.groups,
			dilation=self.dilation,
			bias=self.bias,
			)(input)
		output = nn.LayerNorm(
			epsilon=self.layer_norm_eps,
			)(output) if self.layer_norm_eps else output
		output = self.act(output)
		return output



"""
Nested transformer (NesT).
"""


import typing as T
from functools import partial
from math import sqrt

from flax import linen as nn
from jax import numpy as jnp

from .. import layers
from .factory import register_configs


class ConvLNPool(nn.Module):
	"""
	Convolution followed by layer normalization and max pooling
	for downsampling.

	Args:
		token_dim (int): Token dimension.
		jax (bool): Whether padding should be fixed at 'same' for compatibility
		with JAX.
		Default is True.
	"""
	token_dim: int
	jax: bool = True

	@nn.compact
	def __call__(self, input):
		output = layers.ConvLNAct(
			out_dim=self.token_dim,
			kernel_size=3,
			padding='same' if self.jax else 1,
			)(input)
		output = layers.max_pool(
			input=output,
			stride=2,
			padding='same' if self.jax else 1,
			)
		return output


class BlockMHSA(nn.Module):
	"""
	Multi-headed self-attention with support for an extra
	block axis.

	Args:
		n_heads (int): Number of heads.
	"""
	n_heads: int

	@nn.compact
	def __call__(self, input):
		bs, n_blocks, n_tokens, token_dim = input.shape
		head_dim = token_dim//self.n_heads

		qkv = nn.Dense(
			features=3*token_dim,
			)(input)
		qkv = jnp.reshape(qkv, (bs, n_blocks, n_tokens, 3, self.n_heads, head_dim))
		qkv = jnp.transpose(qkv, (3, 0, 4, 1, 2, 5))
		q, k, v = jnp.split(
			ary=qkv,
			indices_or_sections=3,
			axis=0,
			)
		q, k, v = jnp.squeeze(q, axis=0), jnp.squeeze(k, axis=0), jnp.squeeze(v, axis=0)

		attention = q @ jnp.swapaxes(k, axis1=-2, axis2=-1) / jnp.sqrt(q.shape[-1])
		attention = nn.softmax(attention)

		output = attention @ v
		output = jnp.transpose(output, (0, 2, 3, 4, 1))
		output = jnp.reshape(output, (bs, n_blocks, n_tokens, token_dim))

		output = nn.Dense(
			features=output.shape[-1],
			)(output)

		return output


def block_partition(
	input,
	block_size: T.Union[T.Tuple[int, int], int],
	):
	"""
	Partitions the input into blocks.

	Args:
		input: Input.
		block_size (T.Union[T.Tuple[int, int], int]): Block size.
		If an int, this value is used along both spatial dimensions.
	
	Returns: The input partitioned into blocks.
	"""
	bs, h, w, in_dim = input.shape
	block_h, block_w = layers.tuplify(block_size)
	n_blocks_h, n_blocks_w = h//block_h, w//block_w

	output = jnp.reshape(input, (bs, n_blocks_h, block_h, n_blocks_w, block_w, in_dim))
	output = jnp.transpose(output, (0, 1, 3, 2, 4, 5))
	output = jnp.reshape(output, (bs, n_blocks_h*n_blocks_w, -1, in_dim))
	return output


def block_merge(
	input,
	block_size: T.Union[T.Tuple[int, int], int],
	):
	"""
	Merges blocks.

	Args:
		input: Input.
		block_size (T.Union[T.Tuple[int, int], int]): Block size.
		If an int, this value is used along both spatial dimensions.
	
	Returns: The input partitioned into blocks.
	"""
	bs, n_blocks, n_tokens, token_dim = input.shape
	block_h, block_w = layers.tuplify(block_size)
	n_blocks_h = n_blocks_w = int(sqrt(n_blocks))

	output = jnp.reshape(input, (bs, n_blocks_h, n_blocks_w, block_h, block_w, token_dim))
	output = jnp.transpose(output, (0, 1, 3, 2, 4, 5))
	output = jnp.reshape(output, (bs, block_h*n_blocks_h, block_w*n_blocks_w, token_dim))
	return output


class NesTStage(nn.Module):
	"""
	NesT stage, i.e., a single hierarchial level.

	Args:
		depth (int): Depth.
		token_dim (int): Token dimension.
		n_heads (int): Number of heads.
		block_size (T.Union[T.Tuple[int, int], int]): Block size.
		If an int, this value is used along both spatial dimensions.
		downsample (bool): Whether to downsample.
		Default is False.
		jax (bool): Whether padding should be fixed at 'same' for compatibility
		with JAX.
		Default is True.
	"""
	depth: int
	token_dim: int
	n_heads: int
	block_size: T.Union[T.Tuple[int, int], int] 
	downsample: bool = False
	jax: bool = True
	
	@nn.compact
	def __call__(self, input):
		if self.downsample:
			input = ConvLNPool(
				token_dim=self.token_dim,
				jax=self.jax,
				)(input)

		output = block_partition(input, self.block_size)
		output = layers.AbsPosEmbed(
			n_axes=-3,
			)(output)
		
		for _ in range(self.depth):
			output = layers.MetaFormerBlock(
				token_mixer=partial(BlockMHSA, n_heads=self.n_heads),
				)(output)
		output = block_merge(output, self.block_size) 

		return output


class NesT(nn.Module):
	"""
	Nested transformer.

	Args:
		depths (T.Tuple[int, ...]): Depth of each stage.
		token_dims (T.Tuple[int, ...]): Token dimension of each stage.
		n_heads (T.Tuple[int, ...]): Number of heads of each
		stage.
		patch_size (int): Patch size. This value is used
		along both spatial dimensions.
		Default is 16.
		jax (bool): Whether padding should be fixed at 'same' for compatibility
		with JAX.
		Default is True.
		n_classes (int): Number of output classes. If 0, there is no head,
		and the raw final features are returned. If -1, all stages of the 
		head, other than the final linear layer, are applied and the output
		returned.
		Default is 0.
	"""
	depths: T.Tuple[int, ...]
	token_dims: T.Tuple[int, ...]
	n_heads: T.Tuple[int, ...]
	patch_size: int = 4
	jax: bool = True
	n_classes: int = 0

	@nn.compact
	def __call__(self, input):
		output = layers.PatchEmbed(
			token_dim=self.token_dims[0],
			patch_size=self.patch_size,
			flatten=False,
			)(input)
		self.sow(
			col='intermediates',
			name='stage_0',
			value=output,
			)

		block_size = output.shape[-2] // int(sqrt(4 ** (len(self.depths)-1)))
		for stage_ind in range(len(self.depths)):
			output = NesTStage(
				depth=self.depths[stage_ind],
				token_dim=self.token_dims[stage_ind],
				n_heads=self.n_heads[stage_ind],
				block_size=block_size,
				downsample=False if stage_ind == 0 else True,
				jax=self.jax,
				)(output)
			self.sow(
				col='intermediates',
				name=f'stage_{stage_ind+1}',
				value=output,
				)
		
		output = layers.Head(
			n_classes=self.n_classes,
			layer_norm_eps=1e-6,
			norm_first=True,
			)(output)
		
		return output


@register_configs
def get_nest_configs() -> T.Tuple[T.Type[NesT], T.Dict]:
	"""
	Gets configurations for all available
	NesT models.

	Returns (T.Tuple[T.Type[NesT], T.Dict]): The NesT class and
	configurations of all available models.
	"""
	configs = {
		'nest_tiny_224': {
			'depths': (2, 2, 8),
			'token_dims': (96, 192, 384),
			'n_heads': (3, 6, 12),
			},
		'nest_small_224': {
			'depths': (2, 2, 20),
			'token_dims': (96, 192, 384),
			'n_heads': (3, 6, 12),
			},
		'nest_base_224': {
			'depths': (2, 2, 20),
			'token_dims': (128, 256, 512),
			'n_heads': (4, 8, 16),
			},
		}
	return NesT, configs



"""
Functions related to shifted window attention by Liu et al.:
- cyclic_shift: Cyclical shift.
- window_partition: Partitions the input into windows.
- window_merge: Merges windows.

References:
- Liu et al. Swin Transformer: Hierarchical Vision Transformer using Shifted Windows.
"""


__all__ = [
	'cyclic_shift',
	'window_partition',
	'window_merge',
	]


import typing as T

from jax import numpy as jnp

from .tuplify import tuplify


def cyclic_shift(
	input,
	shift_size: T.Union[T.Tuple[int, int], int],
	):
	"""
	Cyclic-shifts the input.

	Args:
		input: Input.
		shift_size (T.Union[T.Tuple[int, int], int]): Shift size.
		If an int, this value is used along both spatial dimensions.
	
	Returns: Cyclic-shifted version of input.
	"""
	shift_size = tuplify(shift_size)
	return jnp.roll(
		input,
		shift=(-shift_size[0], -shift_size[1]),
		axis=(-3, -2),
		)


def window_partition(
	input,
	window_size: T.Union[T.Tuple[int, int], int],
	):
	"""
	Partitions the input into windows.

	Args:
		input: Input.
		window_size (T.Union[T.Tuple[int, int], int]): Window size.
		If an int, this value is used along both spatial dimensions.
	
	Returns: The input partitioned into windows.
	"""
	bs, h, w, in_dim = input.shape
	window_h, window_w = tuplify(window_size)

	output = jnp.reshape(input, (bs, h//window_h, window_h, w//window_w, window_w, in_dim))
	output = jnp.transpose(output, (0, 1, 3, 2, 4, 5))
	output = jnp.reshape(output, (-1, window_h, window_w, in_dim))
	return output


def window_merge(
	input,
	img_size: T.Union[T.Tuple[int, int], int],
	window_size: T.Union[T.Tuple[int, int], int],
	):
	"""
	Merges windows.

	Args:
		input: Input.
		img_size (T.Union[T.Tuple[int, int], int]): Image size.
		If an int, this value is used along both spatial dimensions.
		window_size (T.Union[T.Tuple[int, int], int]): Window size.
		If an int, this value is used along both spatial dimensions.
	
	Returns: The merged version of the input's windows.
	"""
	c = input.shape[-1]
	img_h, img_w = tuplify(img_size)
	window_h, window_w = tuplify(window_size)

	output = jnp.reshape(input, (-1, img_h//window_h, img_w//window_w, window_h, window_w, c))
	output = jnp.transpose(output, (0, 1, 3, 2, 4, 5))
	output = jnp.reshape(output, (-1, img_h, img_w, c))
	return output



"""
Functions for accessing flaim models and loading pre-trained parameters.
"""


import os
import pickle
import re
import urllib.request
import typing as T

import jax
from flax import traverse_util
from flax import linen as nn
from jax import numpy as jnp


def add_model_cls(
	model_cls: T.Callable,
	configs: T.Dict,
	) -> T.Dict:
	"""
	Adds the desired model class to every element in a
	dictionary of model configurations.

	Args:
		model_cls (T.Callable): Model class.
		configs (T.Dict): Dictionary of model configurations.
	
	Returns (T.Dict): Model configurations, with the model class
	added to them.
	"""
	for model_name in configs:
		configs[model_name]['model_cls'] = model_cls
	return configs


CONFIGS = {}
def register_configs(
	get_configs: T.Callable,
	) -> T.Callable:
	"""
	Decorator for registering model configurations
	by adding them to CONFIGS.

	Args:
		get_configs (T.Callable): T.Callable that returns 
		model configurations.
	
	Returns (T.Callable): get_configs unmodified.
	"""
	model_cls, configs = get_configs()
	configs = add_model_cls(model_cls, configs)
	CONFIGS.update(configs)
	return get_configs


def get_input_size(
	model_name: str,
	) -> int:
	"""
	Gets the input size a model expects.

	Args:
		model_name (str): Name of model.
	
	Returns (int): The input size the model expects.
	"""
	size = re.findall(r'_\d{3}(?:_|$)', model_name)
	return int(size[0].split('_')[1]) if size else 224


def create_flaim_dir() -> str:
	"""
	Creates a .flaim/ folder in the home directory if it doesn't
	already exist.

	Returns (str): Path to the .flaim/ directory
	"""
	home_dir = os.path.expanduser("~")
	flaim_dir = home_dir+'/.flaim/'

	if not os.path.exists(flaim_dir):
		os.mkdir(flaim_dir)
	
	return flaim_dir


def download_vars(
	model_name: str,
	) -> str:
	"""
	Downloads a model's parameters if it has not already been downloaded
	and saves it to the .flaim/ directory.

	Args:
		model_name (str): Name of model whose parameters are to be downloaded.
	
	Returns (str): Path to the parameters' file.
	"""
	flaim_dir = create_flaim_dir()
	params_path = flaim_dir+model_name
	
	if not os.path.exists(params_path):
		url = f'https://huggingface.co/BobMcDear/{model_name}/resolve/main/{model_name}.pickle'
		urllib.request.urlretrieve(
			url=url,
			filename=params_path,
			)
	
	return params_path


def load_pretrained_vars(
	model_name: str,
	) -> T.Dict:
	"""
	Loads a model's parameters from the .flaim/ directory. If it doesn't exist,
	it is downloaded.

	Args:
		model_name (str): Name of model whose parameters are returned.
	
	Returns (T.Dict): The model's parameters.
	"""
	params_path = download_vars(model_name)
	with open(params_path, 'rb') as f:
		params = pickle.load(f)
	return params


def init_model(
	model: nn.Module,
	jit: bool = True,
	prng: T.Optional[jax.random.KeyArray] = None,
	input_size: int = 224,
	):
	"""
	Initializes a model's parameters.

	Args:
		model (nn.Module): Model to initialize.
		jit (bool): Whether to JIT the initialization function.
		Default is True.
		prng (T.Optional[jax.random.KeyArray]): PRNG key for
		initializing the model. If None, a key is created.
		Default is None.
		input_size (int): The input size the model expects.
		Default is 224.
	"""
	init_fn = jax.jit(model.init) if jit else model.init
	return init_fn(prng or jax.random.PRNGKey(0), jnp.ones((1, input_size, input_size, 3)))


def merge_vars(
	vars: T.Dict,
	pretrained_vars: T.Dict,
	) -> T.Dict:
	"""
	Merges dictionaries of randomly-initialized and pre-trained variables,
	with the pre-trained variables taking precedence in case of duplicates.

	Args:
		vars (T.Dict): Randomly-initialized variables.
		pretrained_vars (T.Dict): Pre-trained variables.
	
	Returns (T.Dict): Merged dictionary of vars and pretrained_vars.
	"""
	vars = traverse_util.flatten_dict(vars)
	pretrained_vars = traverse_util.flatten_dict(pretrained_vars)

	merged_vars = {
		**vars,
		**pretrained_vars,
		} 
	merged_vars = traverse_util.unflatten_dict(merged_vars)

	return merged_vars


def get_model(
	model_name: str,
	pretrained: bool = True,
	n_classes: int = 0,
	jit: bool = True,
	prng: T.Optional[jax.random.KeyArray] = None,
	) -> T.Tuple[nn.Module, T.Dict]:
	"""
	Returns a model and its parameters.

	Args:
		model_name (str): Name of model.
		pretrained (bool): Whether to return pre-trained parameters.
		If False, the parameters are randomly initialized.
		Default is True.
		n_classes (int): Number of output classes. If 0, there is no head,
		and the raw final features are returned. If -1, all stages of the 
		head, other than the final linear layer, are applied and the output
		returned.
		Default is 0.
		jit (bool): Whether to JIT the initialization function.
		Default is True.
		prng (T.Optional[jax.random.KeyArray]): PRNG key for
		initializing the model's parameters. If None, a key is created.
		Default is None.

	Returns (T.Tuple[nn.Module, T.Dict]): The model and its parameters.
	"""
	if model_name not in CONFIGS:
		raise ValueError(f'{model_name} is not a recognized model.')

	config = CONFIGS[model_name]
	config['n_classes'] = n_classes

	model_cls = config.pop('model_cls')
	input_size = get_input_size(model_name)

	model = model_cls(**config)
	config['model_cls'] = model_cls

	vars = init_model(
		model=model,
		jit=jit,
		prng=prng,
		input_size=input_size,
		)
	
	if pretrained:
		pretrained_vars = load_pretrained_vars(model_name)
		vars = merge_vars(vars, pretrained_vars)
	
	return model, vars


def list_models(
	pattern: T.Optional[str] = None,
	) -> T.List:
	"""
	Lists available models.

	Args:
		pattern (T.Optional[str]): If not None, only the names of
		the models that contain this regex pattern are returned.
		Default is None.

	Returns (T.List): List of names of all available models.
	"""
	names = list(CONFIGS)
	return [name for name in names if bool(re.search(pattern, name))] if pattern else names



"""
Visual attention network (VAN).
"""


import typing as T

from flax import linen as nn

from .. import layers
from .factory import register_configs


class VANAttention(nn.Module):
	"""
	VAN attention module.

	Args:
		dw_kernel_size (int): Kernel size of
		the depthwise convolution.
		Default is 5.
		dilated_dw_kernel_size (int): Kernel size of
		the dilated depthwise convolution.
		Default is 7.
	"""
	dw_kernel_size: int = 5
	dilated_dw_kernel_size: int = 7

	@nn.compact
	def __call__(self, input):
		output = layers.Conv(
			kernel_size=1,
			)(input)
		output = layers.gelu(output)
		output = layers.LKA(
			dw_kernel_size=self.dw_kernel_size,
			dilated_dw_kernel_size=self.dilated_dw_kernel_size,
			)(output)
		output = layers.Conv(
			kernel_size=1,
			)(output)
		return input+output


class VANBlock(nn.Module):
	"""
	VAN block.

	Args:
		mlp_hidden_dim_expansion_factor (float): Factor of expansion
		for the hidden neurons of the MLP.
		Default is 4.
		dw_kernel_size (int): Kernel size of the depthwise convolution for LKA.
		Default is 5.
		dilated_dw_kernel_size (int): Kernel size of the dilated convolution for LKA.
		Default is 7.
		layer_scale_init_value (T.Optional[float]): Value
		for initializing LayerScale. If None, LayerScale is not
		applied.
		Default is 1e-2.
	"""
	mlp_hidden_dim_expansion_factor: float = 4.
	dw_kernel_size: int = 5
	dilated_dw_kernel_size: int = 7
	layer_scale_init_value: T.Optional[float] = 1e-2

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = nn.BatchNorm(
			use_running_average=not training,
			)(input)
		output = VANAttention(
			dw_kernel_size=self.dw_kernel_size,
			dilated_dw_kernel_size=self.dilated_dw_kernel_size,
			)(output)
		output = layers.LayerScale(
			init_value=self.layer_scale_init_value,
			)(output)
		output = input+output

		residual = output
		output = nn.BatchNorm(
			use_running_average=not training,
			)(output)
		output = layers.TransformerMLP(
			hidden_dim_expansion_factor=self.mlp_hidden_dim_expansion_factor,
			dw_kernel_size=3,
			layer_norm_eps=None,
			layer_scale_init_value=self.layer_scale_init_value,
			residual=False,
			)(output)
		output = residual+output

		return output


class VANStage(nn.Module):
	"""
	VAN stage.

	Args:
		depth (int): Depth.
		out_dim (int): Number of output channels.
		mlp_hidden_dim_expansion_factor (float): Factor of expansion
		for the hidden neurons of the MLP.
		Default is 4.
		dw_kernel_size (int): Kernel size of the depthwise convolution for LKA.
		Default is 5.
		dilated_dw_kernel_size (int): Kernel size of the dilated depthwise convolution
		for LKA.
		Default is 7.
		layer_scale_init_value (T.Optional[float]): Value
		for initializing LayerScale. If None, LayerScale is not
		applied.
		downsample (bool): Whether there should be a downsampling
		module at the beginning.
		Default is False.
	"""
	depth: int
	out_dim: int
	mlp_hidden_dim_expansion_factor: float = 4.
	dw_kernel_size: int = 5
	dilated_dw_kernel_size: int = 7
	layer_scale_init_value: T.Optional[float] = 1e-2
	downsample: bool = False

	@nn.compact
	def __call__(self, input, training: bool = True):
		if self.downsample:
			input = layers.ConvBNAct(
				out_dim=self.out_dim,
				stride=2,
				bias_force=True,
				)(input, training=training)

		for _ in range(self.depth):
			input = VANBlock(
				dw_kernel_size=self.dw_kernel_size,
				dilated_dw_kernel_size=self.dilated_dw_kernel_size,
				mlp_hidden_dim_expansion_factor=self.mlp_hidden_dim_expansion_factor,
				layer_scale_init_value=self.layer_scale_init_value,
				)(input, training=training)
		output = nn.LayerNorm()(input)

		return output


class VAN(nn.Module):
	"""
	Visual attention network.

	Args:
		depths (T.Tuple[int, ...]): Depth of each stage.
		out_dims (T.Tuple[int, ...]): Number of output channels of each stage.
		mlp_hidden_dim_expansion_factor (T.Tuple[float, ...]: Factor of expansion
		for the hidden neurons of the MLP of each stage.
		Default is (8, 8, 4, 4).
		dw_kernel_size (int): Kernel size of the depthwise convolution for LKA.
		Default is 5.
		dilated_dw_kernel_size (int): Kernel size of the dilated depthwise convolution
		for LKA.
		Default is 7.
		layer_scale_init_value (T.Optional[float]): Value
		for initializing LayerScale. If None, LayerScale is not
		applied.
		Default is 1e-2.
		n_classes (int): Number of output classes. If 0, there is no head,
		and the raw final features are returned. If -1, all stages of the 
		head, other than the final linear layer, are applied and the output
		returned.
		Default is 0.
	"""
	depths: T.Tuple[int, ...]
	out_dims: T.Tuple[int, ...]
	mlp_hidden_dim_expansion_factors: T.Tuple[float, ...] = (8., 8., 4., 4.)
	dw_kernel_size: int = 5
	dilated_dw_kernel_size: int = 7
	layer_scale_init_value: T.Optional[float] = 1e-2
	n_classes: int = 0

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = layers.ConvBNAct(
			out_dim=self.out_dims[0],
			kernel_size=7,
			stride=4,
			bias_force=True,
			)(input, training=training)
		self.sow(
			col='intermediates',
			name='stage_0',
			value=output,
			)
		
		for stage_ind in range(len(self.depths)):
			output = VANStage(
				depth=self.depths[stage_ind],
				out_dim=self.out_dims[stage_ind],
				mlp_hidden_dim_expansion_factor=self.mlp_hidden_dim_expansion_factors[stage_ind],
				layer_scale_init_value=self.layer_scale_init_value,
				downsample=False if stage_ind == 0 else True,
				)(output, training=training)
			self.sow(
				col='intermediates',
				name=f'stage_{stage_ind+1}',
				value=output,
				)
		
		output = layers.Head(
			n_classes=self.n_classes,
			)(output)

		return output


@register_configs
def get_van_configs() -> T.Tuple[T.Type[VAN], T.Dict]:
	"""
	Gets configurations for all available
	VAN models.

	Returns (T.Tuple[T.Type[VAN], T.Dict]): The VAN class and
	configurations of all available models.
	"""
	configs = {
		'van_b0': {
			'depths': (3, 3, 5, 2),
			'out_dims': (32, 64, 160, 256),
			},
		'van_b1': {
			'depths': (2, 2, 4, 2),
			'out_dims': (64, 128, 320, 512),
			},
		'van_b2': {
			'depths': (3, 3, 12, 3),
			'out_dims': (64, 128, 320, 512),
			},
		'van_b3': {
			'depths': (3, 5, 27, 3),
			'out_dims': (64, 128, 320, 512),
			},
		}
	return VAN, configs



"""
Identities:
- identity: Identity (function).
- Identity: Identity (class).
"""


__all__ = [
	'identity',
	'Identity',
	]


from flax import linen as nn


def identity(input, *args, **kwargs):
	"""
	Identity function.

	Args:
		input: Input.
		*args.
		**kwargs.
	
	Returns: Input.
	"""
	return input


class Identity(nn.Module):
	"""
	Identity class.
	"""
	@nn.compact
	def __call__(self, input, *args, **kwargs):
		return input



"""
Pooling methods:
- avg_pool: Pooling similar to Flax's average pooling but can automatically calulate
padding, accepts integer kernel size, etc.
- max_pool: Pooling similar to Flax's average pooling but can automatically calulate
padding, accepts integer kernel size, etc.
- global_avg_pool: Global average pooling.
- global_max_pool: Global max pooling.
- global_avg_max_pool: Returns a tuple of the globally average- and max-pooled input.
- global_concat_avg_max_pool: Concatenates the results of global 
average and max pooling, by Wightman.
- global_sum_avg_max_pool: Sums the results of global 
average and max pooling, by Wightman.

References:
- Wightman. PyTorch Image Models.
"""


__all__ = [
	'avg_pool',
	'max_pool',
	'global_avg_pool',
	'global_max_pool',
	'global_avg_max_pool',
	'global_concat_avg_max_pool',
	'global_sum_avg_max_pool',
	]


import typing as T

from flax import linen as nn
from jax import numpy as jnp

from .conv import get_kernel_size_stride_padding


def avg_pool(
	input,
	kernel_size: T.Union[T.Tuple[int, int], int] = 3,
	stride: T.Union[T.Tuple[int, int], int] = 1,
	padding: T.Optional[T.Union[str, int]] = None,
	count_include_pad: bool = True,
	):
	"""
	Similar to Flax's average pooling but accepts integer kernel size, 
	supports depthwise convolution, etc.

	Args:
		input: Input.
		kernel_size (T.Union[T.Tuple[int, int], int]): Kernel size.
		If an int, this value is used along both spatial dimensions.
		Default is 3.
		stride (T.Union[T.Tuple[int, int], int]): Stride. If an int,
		this value is used along both spatial dimensions.
		Default is 1.
		padding (T.Optional[T.Union[str, int]]): Padding. If None,
		it is set so the spatial dimensions are exactly divided by stride.
		If an int, this value is used along both spatial dimensions.
		Default is None.
		count_include_pad (bool): Whether to include padding when
		calculating averages.
		Default is True.
	"""
	kernel_size, stride, padding = get_kernel_size_stride_padding(
		kernel_size=kernel_size,
		stride=stride,
		padding=padding,
		)
	return nn.avg_pool(
		inputs=input,
		window_shape=kernel_size,
		strides=stride,
		padding=padding,
		count_include_pad=count_include_pad,
		)


def max_pool(
	input,
	kernel_size: T.Union[T.Tuple[int, int], int] = 3,
	stride: T.Union[T.Tuple[int, int], int] = 1,
	padding: T.Optional[T.Union[str, int]] = None,
	):
	"""
	Similar to Flax's max pooling but accepts integer kernel size, 
	supports depthwise convolution, etc.

	Args:
		input: Input.
		kernel_size (T.Union[T.Tuple[int, int], int]): Kernel size.
		If an int, this value is used along both spatial dimensions.
		Default is 3.
		stride (T.Union[T.Tuple[int, int], int]): Stride. If an int,
		this value is used along both spatial dimensions.
		Default is 1.
		padding (T.Optional[T.Union[str, int]]): Padding. If None,
		it is set so the spatial dimensions are exactly divided by stride.
		If an int, this value is used along both spatial dimensions.
		Default is None.
	"""
	kernel_size, stride, padding = get_kernel_size_stride_padding(
		kernel_size=kernel_size,
		stride=stride,
		padding=padding,
		)
	return nn.max_pool(
		inputs=input,
		window_shape=kernel_size,
		strides=stride,
		padding=padding,
		)


def global_avg_pool(
	input,
	retain_spatial: bool = True,
	):
	"""
	Global  average pooling.

	Args:
		input: Input.
		retain_spatial (bool): Whether the spatial
		dimensions should be retained or squeezed.
		Default is True.
	
	Returns: Globally average-pooled input.
	"""
	return jnp.mean(input, axis=(-3, -2), keepdims=retain_spatial)


def global_max_pool(
	input,
	retain_spatial: bool = True,
	):
	"""
	Global  max pooling.

	Args:
		input: Input.
		retain_spatial (bool): Whether the spatial
		dimensions should be retained or squeezed.
		Default is True.
	
	Returns: Globally average-pooled input.
	"""
	return jnp.max(input, axis=(-3, -2), keepdims=retain_spatial)


def global_avg_max_pool(
	input,
	retain_spatial: bool = True,
	):
	"""
	Returns a tuple of the globally average- and max-pooled input.

	Args:
		input: Input.
		retain_spatial (bool): Whether the spatial
		dimensions should be retained or squeezed.
		Default is True.
	
	Returns: Tuple of the globally average- and max-pooled input.
	"""
	return (
		global_avg_pool(input, retain_spatial),
		global_max_pool(input, retain_spatial),
		)


def global_concat_avg_max_pool(
	input,
	retain_spatial: bool = True,
	):
	"""
	Concatenates the results of global  average and max pooling.

	Args:
		retain_spatial (bool): Whether the spatial
		dimensions should be retained or squeezed.
		Default is True.
	
	Returns: Concatenation of the globally average- and max-pooled input.
	"""
	return jnp.concatenate(global_avg_max_pool(input, retain_spatial), axis=-1)


def global_sum_avg_max_pool(
	input,
	retain_spatial: bool = True,
	):
	"""
	Sums the results of global  average and max pooling.

	Args:
		retain_spatial (bool): Whether the spatial
		dimensions should be retained or squeezed.
		Default is True.
	
	Returns: Summation of the globally average- and max-pooled input.
	"""
	return jnp.sum(global_avg_max_pool(input, retain_spatial), axis=-1)



"""
State-of-the-art pre-trained vision models.
"""


from . import (
	cait,
	convmixer,
	convnext,
	efficientnetv2,
	gcvit,
	hornet,
	maxvit,
	nest,
	pvtv2,
	resnet,
	swin,
	van,
	vgg,
	vit,
	)
from .factory import get_model, list_models



"""
Multi-axis vision transformer (MaxViT).
"""


import typing as T
from functools import partial

import jax
from flax import linen as nn
from jax import numpy as jnp

from .. import layers
from .factory import register_configs


class MaxViTStem(nn.Module):
	"""
	MaxViT stem.

	Args:
		out_dim (int): Number of output channels.
		Default is 64.
		tf (bool): Whether to use padding of 'same' and an
		approximation of GELU for compatibility with TensorFlow.
		Default is True.
	"""
	out_dim: int = 64
	tf: bool = True

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = layers.ConvBNAct(
			out_dim=self.out_dim,
			stride=2,
			bias_force=True,
			act=nn.gelu if self.tf else layers.gelu,
			tf=self.tf,
			)(input, training=training)
		output = layers.Conv(
			padding='same' if self.tf else 1,
			)(output)
		return output


class PreNormMBConvDownsample(nn.Module):
	"""
	PreNormMBConv downsampling module.

	Args:
		out_dim (int): Number of output channels.
		stride (int): Stride.
		Default is 1.
	"""
	out_dim: int
	stride: int = 1

	@nn.compact
	def __call__(self, input):
		if self.stride != 1:
			input = layers.avg_pool(
				input=input,
				kernel_size=self.stride,
				stride=self.stride,
				padding=0,
				)
		
		if input.shape[-1] != self.out_dim:
			input = layers.Conv(
				out_dim=self.out_dim,
				kernel_size=1,
				)(input)
			
		return input


class PreNormMBConv(nn.Module):
	"""
	MBConv with normalization at the beginning.

	Args:
		out_dim (int): Number of output channels.
		stride (int): Stride.
		Default is 1.
		expansion_factor (int): Expansion factor for
		the inverted bottleneck.
		Default is 4.
		act (T.Callable): Activation function.
		Default is nn.gelu.
		tf (bool): Whether to use batch normalization epsilon of
		1e-3 and padding of 'same' for compatibility with TensorFlow.
		Default is True.
	"""
	out_dim: int
	stride: int = 1
	expansion_factor: int = 4
	act: T.Callable = nn.gelu
	tf: bool = True

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = nn.BatchNorm(
			use_running_average=not training,
			epsilon=1e-3 if self.tf else 1e-5,
			)(input)

		output = layers.ConvBNAct(
			out_dim=4*self.out_dim,
			kernel_size=1,
			act=self.act,
			tf=self.tf,
			)(output, training=training)
		output = layers.ConvBNAct(
			stride=self.stride,
			groups='dw',
			act=self.act,
			tf=self.tf,
			)(output, training=training)
		output = layers.SE(
			reduction_factor=self.expansion_factor*4,
			act=nn.silu,
			)(output)
		output = layers.Conv(
			out_dim=self.out_dim,
			kernel_size=1,
			)(output)
		
		return output + PreNormMBConvDownsample(
			out_dim=self.out_dim,
			stride=self.stride,
			)(input)


def grid_partition(
	input,
	grid_size: T.Union[T.Tuple[int, int], int],
	):
	"""
	Partitions the input into grids.

	Args:
		input: Input.
		grid_size (T.Union[T.Tuple[int, int], int]): Grid size.
		If an int, this value is used along both spatial dimensions.
	
	Returns: The input partitioned into grids.
	"""
	bs, h, w, in_dim = input.shape
	grid_h, grid_w = layers.tuplify(grid_size)
	n_grids_h, n_grids_w = h//grid_h, w//grid_w

	output = jnp.reshape(input, (bs, grid_h, n_grids_h, grid_w, n_grids_w, in_dim))
	output = jnp.transpose(output, (0, 2, 4, 1, 3, 5))
	output = jnp.reshape(output, (-1, grid_h, grid_w, in_dim))
	return output


def grid_merge(
	input,
	img_size: T.Union[T.Tuple[int, int], int],
	grid_size: T.Union[T.Tuple[int, int], int],
	):
	"""
	Merges grids.

	Args:
		input: Input.
		img_size (T.Union[T.Tuple[int, int], int]): Image size.
		If an int, this value is used along both spatial dimensions.
		grid_size (T.Union[T.Tuple[int, int], int]): Grid size.
		If an int, this value is used along both spatial dimensions.
	
	Returns: The merged version of the input's grids.
	"""
	in_dim = input.shape[-1]
	img_h, img_w = layers.tuplify(img_size)
	grid_h, grid_w = layers.tuplify(grid_size)

	output = jnp.reshape(input, (-1, img_h//grid_h, img_w//grid_w, grid_h, grid_w, in_dim))
	output = jnp.transpose(output, (0, 3, 1, 4, 2, 5))
	output = jnp.reshape(output, (-1, img_h, img_w, in_dim))
	return output


def get_maxvit_rel_pos_ind(
	window_size: int,
	) -> jnp.ndarray:
	"""
	Gets a matrix used to index MaxViT's relative position bias
	table.

	Args:
		window_size (int): Window size. This value is used along
		both spatial dimesnoins.
	
	Returns (jnp.ndarray): Matrix used to index MaxViT's relative position bias
	table.
	"""
	max_rel_pos = window_size-1
	n_rel_distance = 2*max_rel_pos + 1

	rel_pos_ind = jnp.zeros((window_size, window_size, n_rel_distance))
	for i in range(window_size):
		for j in range(window_size):
			k = j - i + max_rel_pos
			if max_rel_pos < abs(j-i):
				continue
			rel_pos_ind = rel_pos_ind.at[i, j, k].set(1)

	return rel_pos_ind


def index_maxvit_rel_pos_table(
	rel_pos_table,
	rel_pos_ind,
	):
	"""
	Indexes a relative position bias table given a matrix of indices 
	for MaxViT.

	Args:
		rel_pos_table: Relative position bias table to index.
		rel_pos_ind: Matrix used to index rel_pos_table.

	Returns: Desired elements of rel_pos_table. 
	"""
	rel_pos_table = jnp.einsum('nhw,ixh->nixw', rel_pos_table, rel_pos_ind)
	rel_pos_table = jnp.einsum('nixw,jyw->nijxy', rel_pos_table, rel_pos_ind)

	area = len(rel_pos_ind) ** 2
	rel_pos_table = jnp.reshape(rel_pos_table, (len(rel_pos_table), area, area))
	return rel_pos_table


class MaxViTRelPosEmbed(nn.Module):
	"""
	Relative position embedding used by MaxViT.

	Args:
		n_heads (int): Number of heads.
		window_size (int): Window size. This value is used along
		both spatial dimesnoins.
	"""
	n_heads: int
	window_size: int

	@nn.compact
	def __call__(self, input):
		size = 2*self.window_size - 1
		rel_pos_table = self.param(
			name='rel_pos_table',
			init_fn=lambda prng: jax.random.normal(prng, (self.n_heads, size, size)),
			)
		rel_pos_ind = self.variable(
			col='rel_pos_ind',
			name='rel_pos_ind',
			init_fn=lambda: get_maxvit_rel_pos_ind(self.window_size),
			).value

		rel_pos_bias = index_maxvit_rel_pos_table(rel_pos_table, rel_pos_ind)
		return input+rel_pos_bias


class PartitionMHSA(nn.Module):
	"""
	Multi-headed self-attention, with support for
	partitioning.

	Args:
		partition_fn (T.Callable): T.Callable used to 
		partition the input.
		merge_fn (T.Callable): T.Callable used to
		merge the data after its partitioning.
		partition_size (int): Size of each partition.
		head_dim (int): Dimension per head.
		Default is 32.
	"""
	partition_fn: T.Callable
	merge_fn: T.Callable
	partition_size: int
	head_dim: int = 32

	@nn.compact
	def __call__(self, input):
		bs, h, w, in_dim = input.shape
		n_heads = in_dim//self.head_dim

		output = self.partition_fn(input, self.partition_size)
		output = jnp.reshape(output, (-1, self.partition_size ** 2, input.shape[-1]))

		output = layers.MHSA(
			to_qkv=n_heads,
			pre_softmax=partial(
				MaxViTRelPosEmbed,
				n_heads=n_heads,
				window_size=self.partition_size,
				),
			)(output)

		output = jnp.reshape(output, (-1, self.partition_size, self.partition_size, in_dim))
		output = self.merge_fn(output, (h, w), self.partition_size)

		return output


class MaxViTBlock(nn.Module):
	"""
	MaxViT block.

	Args:
		out_dim (int): Number of output channels.
		window_size (int): Window size for relative position
		embedding and window attention. This value is used
		along both spatial dimensions.
		grid_size (int): Grid size for relative position
		embedding and grid attention. This value is used
		along both spatial dimensions.
		stride (int): Stride.
		Default is 1.
		head_dim (int): Dimension per head.
		Default is 32.
		tf (bool): Whether to use batch normalization epsilon of
		1e-3, layer normalization epsilon of 1e-5, padding of 'same',
		and an approximation of GELU for compatibility with TensorFlow.
		Default is True.
	"""
	out_dim: int
	window_size: int
	grid_size: int
	stride: int = 1
	head_dim: int = 32
	tf: bool = True

	@nn.compact
	def __call__(self, input, training: bool = True):
		act = nn.gelu if self.tf else layers.gelu

		output = PreNormMBConv(
			out_dim=self.out_dim,
			stride=self.stride,
			act=act,
			tf=self.tf,
			)(input, training=training)
		output = layers.MetaFormerBlock(
			token_mixer=partial(
				PartitionMHSA,
				partition_fn=layers.window_partition,
				merge_fn=layers.window_merge,
				partition_size=self.window_size,
				head_dim=self.head_dim,
				),
			act=act,
			layer_norm_eps=1e-5 if self.tf else 1e-6,
			)(output)
		output = layers.MetaFormerBlock(
			token_mixer=partial(
				PartitionMHSA,
				partition_fn=grid_partition,
				merge_fn=grid_merge,
				partition_size=self.window_size,
				head_dim=self.head_dim,
				),
			act=act,
			layer_norm_eps=1e-5 if self.tf else 1e-6,
			)(output)

		return output


class MaxViTStage(nn.Module):
	"""
	MaxViT stage.

	Args:
		depth (int): Depth.
		out_dim (int): Number of output channels.
		window_size (int): Window size for relative position
		embedding and window attention. This value is used
		along both spatial dimensions.
		grid_size (int): Grid size for relative position
		embedding and grid attention. This value is used
		along both spatial dimensions.
		stride (int): Stride.
		Default is 1.
		head_dim (int): Dimension per head.
		Default is 32.
		tf (bool): Whether to use batch normalization epsilon of
		1e-3, layer normalization epsilon of 1e-5, padding of 'same',
		and an approximation of GELU for compatibility with TensorFlow.
		Default is True.
	"""
	depth: int
	out_dim: int
	window_size: int
	grid_size: int
	stride: int = 1
	head_dim: int = 32
	tf: bool = True
	
	@nn.compact
	def __call__(self, input, training: bool = False):
		for block_ind in range(self.depth):
			input = MaxViTBlock(
				out_dim=self.out_dim,
				stride=self.stride if block_ind == 0 else 1,
				window_size=self.window_size,
				grid_size=self.grid_size,
				head_dim=self.head_dim,
				tf=self.tf,
				)(input, training=training)
		return input


class MaxViTHead(nn.Module):
	"""
	MaxViT head.

	Args:
		n_classes (int): Number of output classes. If 0, the input is returned.
		If -1, all stages of the head, other than the final linear layer,
		are applied and the output returned. 
		Default is 0.
		tf (bool): Whether to use batch normalization epsilon of
		1e-3, layer normalization epsilon of 1e-5, padding of 'same',
		and an approximation of GELU for compatibility with TensorFlow.
		Default is True.
	"""
	n_classes: int = 0
	tf: bool = True

	@nn.compact
	def __call__(self, input):
		if self.n_classes == 0:
			return input

		output = layers.Head(
			n_classes=input.shape[-1],
			layer_norm_eps=1e-5 if self.tf else 1e-6,
			)(input)
		output = nn.tanh(output)
		
		if self.n_classes != -1:
			output = nn.Dense(
				features=self.n_classes,
				)(output)
		
		return output


class MaxViT(nn.Module):
	"""
	Multi-axis vision transformer.

	Args:
		depths (T.Tuple[int, ...]): Depth of each stage.
		out_dims (T.Tuple[int, ...]): Number of output channels of each stage.
		stem_out_dim (int): Number of output channels of the stem.
		Default is 64.
		partition_size_factor (int): Factor by which the spatial dimensions
		of the input are divided to obtain the partition size for window
		and grid attention.
		Default is 32.
		head_dim (int): Dimension per head.
		Default is 32.
		tf (bool): Whether to use batch normalization epsilon of
		1e-3, layer normalization epsilon of 1e-5, padding of 'same',
		and an approximation of GELU for compatibility with TensorFlow.
		Default is True.
		n_classes (int): Number of output classes. If 0, there is no head,
		and the raw final features are returned. If -1, all stages of the 
		head, other than the final linear layer, are applied and the output
		returned.
		Default is 0.
	"""
	depths: T.Tuple[int, ...]
	out_dims: T.Tuple[int, ...]
	stem_out_dim: int = 64
	partition_size_factor: int = 32
	head_dim: int = 32
	tf: bool = True
	n_classes: int = 0

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = MaxViTStem(
			out_dim=self.stem_out_dim,
			)(input, training=training)
		self.sow(
			col='intermediates',
			name='stage_0',
			value=output,
			)

		partition_size = input.shape[-2]//self.partition_size_factor
		for stage_ind in range(len(self.depths)):
			output = MaxViTStage(
				depth=self.depths[stage_ind],
				out_dim=self.out_dims[stage_ind],
				stride=2,
				window_size=partition_size,
				grid_size=partition_size,
				head_dim=self.head_dim,
				tf=self.tf,
				)(output, training=training)
			self.sow(
				col='intermediates',
				name=f'stage_{stage_ind+1}',
				value=output,
				)
		
		output = MaxViTHead(
			n_classes=self.n_classes,
			)(output)
		
		return output


@register_configs
def get_maxvit_configs() -> T.Tuple[T.Type[MaxViT], T.Dict]:
	"""
	Gets configurations for all available
	MaxViT models.

	Returns (T.Tuple[T.Type[MaxViT], T.Dict]): The MaxViT class and
	configurations of all available models.
	"""
	configs = {
		'maxvit_tiny_224': {
			'depths': (2, 2, 5, 2),
			'out_dims': (64, 128, 256, 512),
			},
		'maxvit_small_224': {
			'depths': (2, 2, 5, 2),
			'out_dims': (96, 192, 384, 768),
			},
		'maxvit_base_224': {
			'depths': (2, 6, 14, 2),
			'out_dims': (96, 192, 384, 768),
			},
		'maxvit_large_224': {
			'depths': (2, 6, 14, 2),
			'out_dims': (128, 256, 512, 1024),
			'stem_out_dim': 128,
			},
		'maxvit_tiny_384': {
			'depths': (2, 2, 5, 2),
			'out_dims': (64, 128, 256, 512),
			},
		'maxvit_small_384': {
			'depths': (2, 2, 5, 2),
			'out_dims': (96, 192, 384, 768),
			},
		'maxvit_base_384': {
			'depths': (2, 6, 14, 2),
			'out_dims': (96, 192, 384, 768),
			},
		'maxvit_large_384': {
			'depths': (2, 6, 14, 2),
			'out_dims': (128, 256, 512, 1024),
			'stem_out_dim': 128,
			},
		'maxvit_tiny_512': {
			'depths': (2, 2, 5, 2),
			'out_dims': (64, 128, 256, 512),
			},
		'maxvit_small_512': {
			'depths': (2, 2, 5, 2),
			'out_dims': (96, 192, 384, 768),
			},
		'maxvit_base_512': {
			'depths': (2, 6, 14, 2),
			'out_dims': (96, 192, 384, 768),
			},
		'maxvit_large_512': {
			'depths': (2, 6, 14, 2),
			'out_dims': (128, 256, 512, 1024),
			'stem_out_dim': 128,
			},
		'maxvit_base_384_in22ft1k': {
			'depths': (2, 6, 14, 2),
			'out_dims': (96, 192, 384, 768),
			},
		'maxvit_large_384_in22ft1k': {
			'depths': (2, 6, 14, 2),
			'out_dims': (128, 256, 512, 1024),
			'stem_out_dim': 128,
			},
		'maxvit_xlarge_384_in22ft1k': {
			'depths': (2, 6, 14, 2),
			'out_dims': (192, 384, 768, 1536),
			'stem_out_dim': 192,
			},
		'maxvit_base_512_in22ft1k': {
			'depths': (2, 6, 14, 2),
			'out_dims': (96, 192, 384, 768),
			},
		'maxvit_large_512_in22ft1k': {
			'depths': (2, 6, 14, 2),
			'out_dims': (128, 256, 512, 1024),
			'stem_out_dim': 128,
			},
		'maxvit_xlarge_512_in22ft1k': {
			'depths': (2, 6, 14, 2),
			'out_dims': (192, 384, 768, 1536),
			'stem_out_dim': 192,
			},
		}
	return MaxViT, configs



"""
Some of the smaller modules used by the vision transformer by Dosovitskiy et al.:
- PatchEmbed: Patch embedding.
- ClassToken: ClassToken module.
- AbsPosEmbed: Absolution position embedding.

References: 
- Dosovitskiy et al. An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale.
"""


__all__ = [
	'PatchEmbed',
	'ClassToken',
	'AbsPosEmbed',
	]


import typing as T

import jax
from flax import linen as nn
from jax import numpy as jnp

from .conv import Conv
from .identity import identity


class PatchEmbed(nn.Module):
	"""
	Patch embedding.

	Args:
		token_dim (T.Optional[int]): Token dimension. If None, it is
		set to the dimensionality of the input.
		Default is None.
		patch_size (T.Union[T.Tuple[int, int], int]): Patch size. If an
		int, this value is used along both spatial dimensions.
		Default is 16.
		patch_stride (T.Optional[T.Union[T.Tuple[int, int], int]]): Stride of each patch. 
		If None, it is the same as the patch size. If an int,
		this value is used along both spatial dimensions.
		Default is None.
		patch_padding (T.Optional[T.Union[str, int]]): Padding for patches. 
		If an int, this value is used along both spatial dimensions. If None
		and patch_size and patch_stride are the same, it is set to 0. If None
		and patch_size and patch_stride are different, it is set so the spatial
		dimensions are exactly divided by patch_stride.
		Default is None.
		bias (bool): Whether the linear transformation should have a bias term.
		Default is True.
		layer_norm_eps (T.Optional[float]): Epsilon value passed to 
		layer normalization. If None, no normalization is applied, and norm_first
		is ignored.
		Default is None.
		norm_first (bool): Whether to apply layer normalization before the 
		linear transformation instead of after.
		Default is False.
		flatten (bool): Whether the output should be flattened along the spatial
		dimensions.
		Default is True.
	"""
	token_dim: T.Optional[int] = None
	patch_size: T.Union[T.Tuple[int, int], int] = 16
	patch_stride: T.Optional[T.Union[T.Tuple[int, int], int]] = None
	patch_padding: T.Optional[T.Union[str, int]] = None
	bias: bool = True
	layer_norm_eps: T.Optional[float] = None
	norm_first: bool = False
	flatten: bool = True

	@nn.compact
	def __call__(self, input):
		token_dim = self.token_dim or input.shape[-1]
		layer_norm = nn.LayerNorm(self.layer_norm_eps) if self.layer_norm_eps else identity
		patch_stride = self.patch_stride or self.patch_size
		patch_padding = 0 if self.patch_size == patch_stride else None
		
		if self.norm_first:
			output = layer_norm(input)
			output = Conv(
				out_dim=token_dim,
				kernel_size=self.patch_size,
				stride=patch_stride,
				padding=patch_padding,
				bias=self.bias,
				)(output)
		
		else:
			output = Conv(
				out_dim=token_dim,
				kernel_size=self.patch_size,
				stride=patch_stride,
				padding=patch_padding,
				bias=self.bias,
				)(input)
			output = layer_norm(output)
		
		if self.flatten:
			output = jnp.reshape(output, (len(input), -1, token_dim))

		return output


class ClassToken(nn.Module):
	"""
	Class token module.

	Args:
		concat (bool): Whether to concatenate the class
		token to the beginning of the input when returning the output.
		If False, only the class token is returned.
		Default is True.
	"""
	concat: bool = True

	@nn.compact
	def __call__(self, input):
		token_dim = input.shape[-1]
		class_token = self.param(
			name='class_token',
			init_fn=lambda prng: jnp.zeros((1, 1, token_dim)),
			)
		class_token = jnp.broadcast_to(class_token, (len(input), 1, token_dim))
		return jnp.concatenate((class_token, input), axis=1) if self.concat else class_token


class AbsPosEmbed(nn.Module):
	"""
	Learnable absolute position embedding.

	Args:
		n_axes (int): Absolute position embedding
		is applied along the last n_axes axes.
		Default is -2.
		add (bool): Whether to add position embedding to the input
		when returning the output. If False, only the position embedding
		is returned.
		Default is True.
	"""
	n_axes: int = -2
	add: bool = True

	@nn.compact
	def __call__(self, input):
		pos_embed = self.param(
			name='pos_embed',
			init_fn=lambda prng: jax.random.normal(prng, input.shape[self.n_axes:])
			)
		return input+pos_embed if self.add else pos_embed



"""
Multi-headed self-attention (MHSA) and some related modules by Vaswani et al.:
- QKV: Query-key-value extractor.
- scaled_dot_product_similarity: Scaled dot-product similarity function.
- ProjOut: MHSA's regular output projection mechanism.
- MHSA: Multi-headed self-attention.

References:
- Vaswani et al. Attention Is All You Need.
"""


__all__ = [
	'QKV',
	'scaled_dot_product_similarity',
	'ProjOut',
	'MHSA',
	]


import typing as T

from flax import linen as nn
from jax import numpy as jnp

from .identity import Identity


class QKV(nn.Module):
	"""
	Exctracts query, key, and value vectors for multi-headed self-attention.

	Args:
		n_heads (int): Number of heads.
		bias (bool): Whether the linear transformations should have
		a bias term. If False, k_bias is ignored.
		Default is True.
		k_bias (bool): Whether the linear transformation
		for obtaining keys should contain a bias term.
		Default is True.
	"""
	n_heads: int
	bias: bool = True
	k_bias: bool = True

	@nn.compact
	def __call__(self, input):
		n_tokens, token_dim = input.shape[-2:]
		head_dim = token_dim//self.n_heads

		if not self.bias:
			qkv = nn.Dense(
				features=3*token_dim,
				use_bias=False,
				)(input)

		elif self.k_bias:
			qkv = nn.Dense(
				features=3*token_dim,
				)(input)

		else:
			q_bias = self.param('q_bias', lambda prng: jnp.zeros((token_dim)))
			k_bias = self.variable('k_bias_', 'k_bias_', lambda: jnp.zeros((token_dim)))
			v_bias = self.param('v_bias', lambda prng: jnp.zeros((token_dim)))
			qkv = nn.Dense(
				features=3*token_dim,
				use_bias=False,
				)(input) + jnp.concatenate([q_bias, k_bias.value, v_bias])

		qkv = jnp.reshape(qkv, (-1, n_tokens, 3, self.n_heads, head_dim))
		qkv = jnp.transpose(qkv, (2, 0, 3, 1, 4))
		q, k, v = jnp.split(
			ary=qkv,
			indices_or_sections=3,
			axis=0,
			)
			
		return jnp.squeeze(q, axis=0), jnp.squeeze(k, axis=0), jnp.squeeze(v, axis=0)


def scaled_dot_product_similarity(q, k):
	"""
	Scaled dot-product similarity.

	Args:
		q: Queries.
		k: Keys.
	
	Returns: Scaled dot-product similarity between the queries and keys.
	"""
	return q @ jnp.swapaxes(k, axis1=-2, axis2=-1) / jnp.sqrt(q.shape[-1])


class ProjOut(nn.Module):
	"""
	MHSA's regular output projection mechanism.
	"""
	@nn.compact
	def __call__(self, input):
		output = jnp.swapaxes(input, axis1=-2, axis2=-3)
		output = jnp.reshape(output, (len(input), input.shape[-2], -1))
		output = nn.Dense(
			features=output.shape[-1],
			)(output)
		return output


class MHSA(nn.Module):
	"""
	Multi-headed self-attention module.

	Args:
		to_qkv (T.Optional[T.Union[T.Callable, int]]): If input is not None,
		to_qkv should return the module used to extract queries, key, and values.
		If an int, the regular QKV extractor is used, and to_qkv is interpreted 
		as the number of heads. If input is None, to_qkv is ignored.
		Default is None.
		similarity_fn (T.Callable): T.Callable used to calculate
		similarities betweek queries and keys.
		Default is scaled_dot_product_similarity.
		pre_softmax (T.Callable): T.Callable returning a module
		executed immediately before softmax for transforming the
		attention values.
		Default is Identity.
		post_softmax (T.Callable): T.Callable returning a module
		executed immediately after softmax for transforming the
		attention values.
		weigh_values (T.Callable): T.Callable used to weigh
		the value vectors given the attention values.
		Default is jnp.matmul.
		proj_out (T.Callable): T.Callable returning a module that
		projects the weighed value vectors to the output dimension.
		Default is ProjOut.
	"""
	to_qkv: T.Optional[T.Union[T.Callable, int]] = None
	similarity_fn: T.Callable = scaled_dot_product_similarity
	pre_softmax: T.Callable = Identity
	post_softmax: T.Callable = Identity
	weigh_values: T.Callable = jnp.matmul
	proj_out: T.Callable = ProjOut

	@nn.compact
	def __call__(self, input=None, q=None, k=None, v=None):
		if input is not None:
			to_qkv = QKV(n_heads=self.to_qkv) if isinstance(self.to_qkv, int) else self.to_qkv()
			q, k, v = to_qkv(input)
		
		attention = self.similarity_fn(q, k)

		attention = self.pre_softmax()(attention)
		attention = nn.softmax(attention)
		attention = self.post_softmax()(attention)

		output = self.weigh_values(attention, v)
		output = self.proj_out()(output)
		
		return output



"""
Shifted window attention transformer (Swin transformer).
"""


import typing as T
from functools import partial

from flax import linen as nn
from jax import numpy as jnp

from .. import layers
from .factory import register_configs


def get_mask(
	img_size: int,
	window_size: int,
	shift: int,
	) -> jnp.ndarray:
	"""
	Gets mask for shifted window atention.

	Args:
		img_size (int): Image size. This value is used
		along both spatial dimensions.
		window_size (int): Window size for relative position
		embedding and window attention. This value is used
		along both spatial dimensions.
		Default is 7.
		shift (int): Shift size for cyclic shift.
	"""
	mask = jnp.zeros((1, img_size, img_size, 1))
	slices = (
		slice(0, -window_size),
		slice(-window_size, -shift),
		slice(-shift, None),
		)
	cnt = 0
	
	for slice_h in slices:
		for slice_w in slices:
			mask = mask.at[:, slice_h, slice_w, :].set(cnt)
			cnt += 1

	mask = layers.window_partition(mask, window_size)
	mask = jnp.reshape(mask, (-1, window_size ** 2))

	mask = jnp.expand_dims(mask, 1) - jnp.expand_dims(mask, 2)
	mask = jnp.where(mask != 0, float(-100.0), float(0.0))

	return mask


class Mask(nn.Module):
	"""
	Masking module for shifted window atention.

	Args:
		img_size (int): Image size. This value is used
		along both spatial dimensions.
		window_size (int): Window size for relative position
		embedding and window attention. This value is used
		along both spatial dimensions.
		shift (int): Shift size for cyclic shift. This value
		is used along both spatial dimensions.
		If 0, no masking transpires.
	"""
	img_size: int
	window_size: int
	shift: int

	@nn.compact
	def __call__(self, input):
		if self.shift == 0:
			return input

		n_heads, n_tokens = input.shape[-3:-1]
		mask = self.variable(
			col='mask',
			name='mask',
			init_fn=lambda: get_mask(
				window_size=self.window_size,
				img_size=self.img_size,
				shift=self.shift,
				),
			).value
	
		output = jnp.reshape(input, (-1, len(mask), n_heads, n_tokens, n_tokens))
		mask = jnp.expand_dims(mask, (0, 2))
		output = jnp.reshape(mask+output, (-1, n_heads, n_tokens, n_tokens))

		return output


class WindowMHSA(nn.Module):
	"""
	Multi-headed self-attention, with support for
	shifted or non-shifted windows.

	Args:
		n_heads (int): Number of heads.
		window_size (int): Window size for relative position
		embedding and window attention. This value is used
		along both spatial dimensions.
		shift (int): Shift size for cyclic shift. This value
		is used along both spatial dimensions. If 0, no shifting transpires.
		Default is None.
	"""
	n_heads: int
	window_size: int
	shift: int

	@nn.compact
	def __call__(self, input):
		n_tokens, token_dim = input.shape[-2:]
		window_size = self.window_size
		shift = self.shift
		img_size = int(jnp.sqrt(n_tokens))

		if img_size <= window_size:
			window_size = img_size
			shift = 0

		output = jnp.reshape(input, (-1, img_size, img_size, token_dim))
		output = layers.cyclic_shift(output, shift)
		output = layers.window_partition(output, window_size)
		output = jnp.reshape(output, (-1, window_size ** 2, token_dim))

		output = layers.MHSA(
			to_qkv=self.n_heads,
			pre_softmax=lambda: nn.Sequential([
				layers.RelPosEmbed(
					n_heads=self.n_heads,
					window_size=window_size,
					class_token=False,
					),
				Mask(
					img_size=img_size,
					window_size=window_size,
					shift=shift,
					),
				])
			)(output)

		output = layers.window_merge(
			output,
			window_size=window_size,
			img_size=img_size,
			)
		output = layers.cyclic_shift(output, -shift)
		output = jnp.reshape(output, input.shape)

		return output


class PatchMerge(nn.Module):
	"""
	Merges patches for downsampling.
	"""
	@nn.compact
	def __call__(self, input):
		n_tokens, token_dim = input.shape[-2:]
		img_size = int(jnp.sqrt(n_tokens))
		
		output = jnp.reshape(input, (-1, img_size, img_size, token_dim))
		s1 = output[:, 0::2, 0::2, :]
		s2 = output[:, 1::2, 0::2, :]
		s3 = output[:, 0::2, 1::2, :]
		s4 = output[:, 1::2, 1::2, :]

		output = jnp.concatenate([s1, s2, s3, s4], -1)
		output = jnp.reshape(output, (len(output), -1, 4*token_dim))

		output = nn.LayerNorm(
			epsilon=1e-5,
			)(output)
		output = nn.Dense(
			features=2*token_dim,
			use_bias=False,
			)(output)

		return output


class SwinStage(nn.Module):
	"""
	Swin stage.

	Args:
		depth (int): Depth.
		n_heads (int): Number of heads.
		window_size (int): Window size for relative position
		embedding and window attention. This value is used
		along both spatial dimensions.
		Default is 7.
		downsample (bool): Whether to downsample.
		Default is False.
	"""
	depth: int
	n_heads: int
	window_size: int = 7
	downsample: bool = False

	@nn.compact
	def __call__(self, input):
		for block_ind in range(self.depth):
			input = layers.MetaFormerBlock(
				token_mixer=partial(
					WindowMHSA,
					n_heads=self.n_heads,
					window_size=self.window_size,
					shift=self.window_size//2 if block_ind%2 else 0,
					),
				layer_norm_eps=1e-5,
				)(input)
		
		if self.downsample:
			input = PatchMerge()(input)

		return input


class Swin(nn.Module):
	"""
	Shifted window attention transformer.

	Args:
		depths (T.Tuple[int, ...]): Depth of each stage.
		token_dim (int): Token dimension.
		n_heads (T.Tuple[int, ...]): Number of heads of each
		stage.
		patch_size (int): Patch size. This value is used along
		both spatial dimensions.
		Default is 4.
		window_size (T.Union[T.Tuple[int, ...], int]): Window size for 
		relative position embedding and window attention, used along
		both spatial dimensions. If an int, this value is used for every stage,
		and if a T.List, it must contain the window size for each stage.
		Default is 7.
		n_classes (int): Number of output classes. If 0, there is no head,
		and the raw final features are returned. If -1, all stages of the 
		head, other than the final linear layer, are applied and the output
		returned.
		Default is 0.
	"""
	depths: T.Tuple[int, ...]
	token_dim: int
	n_heads: T.Tuple[int, ...]
	patch_size: int = 4
	window_size: T.Union[T.Tuple[int, ...], int] = 7
	n_classes: int = 0

	@nn.compact
	def __call__(self, input):
		output = layers.PatchEmbed(
			token_dim=self.token_dim,
			patch_size=self.patch_size,
			layer_norm_eps=1e-5,
			)(input)
		self.sow(
			col='intermediates',
			name='stage_0',
			value=output,
			)
		
		window_size = layers.tuplify(self.window_size, len(self.depths))
		for stage_ind in range(len(self.depths)):
			output = SwinStage(
				depth=self.depths[stage_ind],
				n_heads=self.n_heads[stage_ind],
				window_size=window_size[stage_ind],
				downsample=True if stage_ind < len(self.depths)-1 else False,
				)(output)
			self.sow(
				col='intermediates',
				name=f'stage_{stage_ind+1}',
				value=output,
				)
		
		output = layers.Head(
			n_classes=self.n_classes,
			layer_norm_eps=1e-5,
			norm_first=True,
			)(output)

		return output


@register_configs
def get_swin_configs() -> T.Tuple[T.Type[Swin], T.Dict]:
	"""
	Gets configurations for all available
	Swin models.

	Returns (T.Tuple[T.Type[Swin], T.Dict]): The Swin class and
	configurations of all available models.
	"""
	configs = {
		'swin_tiny_window7_224': {
			'depths': (2, 2, 6, 2),
			'token_dim': 96,
			'n_heads': (3, 6, 12, 24),
			},
		'swin_small_window7_224': {
			'depths': (2, 2, 18, 2),
			'token_dim': 96,
			'n_heads': (3, 6, 12, 24),
			},
		'swin_base_window7_224': {
			'depths': (2, 2, 18, 2),
			'token_dim': 128,
			'n_heads': (4, 8, 16, 32),
			},
		'swin_large_window7_224': {
			'depths': (2, 2, 18, 2),
			'token_dim': 192,
			'n_heads': (6, 12, 24, 48),
			},
		'swin_base_window12_384': {
			'depths': (2, 2, 18, 2),
			'token_dim': 128,
			'n_heads': (4, 8, 16, 32),
			'window_size': 12,
			},
		'swin_large_window12_384': {
			'depths': (2, 2, 18, 2),
			'token_dim': 192,
			'n_heads': (6, 12, 24, 48),
			'window_size': 12,
			},
		'swin_base_window7_224_in22k': {
			'depths': (2, 2, 18, 2),
			'token_dim': 128,
			'n_heads': (4, 8, 16, 32),
			'window_size': 7,
			},
		'swin_large_window7_224_in22k': {
			'depths': (2, 2, 18, 2),
			'token_dim': 192,
			'n_heads': (6, 12, 24, 48),
			'window_size': 7,
			},
		'swin_base_window12_384_in22k': {
			'depths': (2, 2, 18, 2),
			'token_dim': 128,
			'n_heads': (4, 8, 16, 32),
			'window_size': 12,
			},
		'swin_large_window12_384_in22k': {
			'depths': (2, 2, 18, 2),
			'token_dim': 192,
			'n_heads': (6, 12, 24, 48),
			'window_size': 12,
			},
		'swin_s3_tiny_224': {
			'depths': (2, 2, 6, 2),
			'token_dim': 96,
			'n_heads': (3, 6, 12, 24),
			'window_size': (7, 7, 14, 7),
			},
		'swin_s3_small_224': {
			'depths': (2, 2, 18, 2),
			'token_dim': 96,
			'n_heads': (3, 6, 12, 24),
			'window_size': (14, 14, 14, 7),
			},
		'swin_s3_base_224': {
			'depths': (2, 2, 30, 2),
			'token_dim': 96,
			'n_heads': (3, 6, 12, 24),
			'window_size': (7, 7, 14, 7),
			},
		}
	return Swin, configs



"""
Recursive gated convolution (g^n convolution) by Rao et al.

References:
- Rao et al. HorNet: Efficient High-Order Spatial Interactions with Recursive Gated Convolutions.
"""


__all__ = [
	'GnConv',
	]


from flax import linen as nn
from jax import numpy as jnp

from .lin_norm_act import Conv


class GnConv(nn.Module):
	"""
	Recursive gated convolution.

	Args:
		out_dim (int): Number of output channels.
		order (int): Order of g^n convolution.
		Default is 5.
		scale (float): Scale factor for the output
		of the depthwise convolution.
		Default is 1/3.
	"""
	out_dim: int
	order: int = 5
	scale: float = 1/3

	@nn.compact
	def __call__(self, input):
		out_dims = [self.out_dim // (2 ** (self.order-1))]
		split_dims = [out_dims[0]]
		for ord_ind in range(self.order-2, -1, -1):
			out_dims.append(self.out_dim // (2 ** ord_ind))
			split_dims.append(split_dims[-1] + out_dims[-1])

		output = Conv(
			out_dim=2*self.out_dim,
			kernel_size=1,
			)(input)
		p0, q0 = jnp.split(
			ary=output,
			indices_or_sections=(split_dims[0],),
			axis=-1,
			)

		ps = self.scale * Conv(
			kernel_size=7,
			groups='dw',
			)(q0)
		ps = jnp.split(
			ary=ps,
			indices_or_sections=split_dims[:-1],
			axis=-1,
			)
		
		output = p0 * ps[0]
		for ord_ind in range(self.order-1):
			output = ps[ord_ind+1] * Conv(
				out_dim=out_dims[ord_ind+1],
				kernel_size=1,
				)(output)
		output = Conv(
			kernel_size=1,
			)(output)
			
		return output



"""
Classification heads:
- Head: Standard head with pooling, layer normalization, and a linear layer.
- ViTHead: Head for ViTs with optional pooling of tokens, layer normalization,
and a linear layer.
"""


__all__ = [
	'Head',
	'ViTHead'
	]


import typing as T

from flax import linen as nn
from jax import numpy as jnp

from .pool import global_avg_pool


class Head(nn.Module):
	"""
	Standard head with pooling, layer normalization, and a linear layer.

	Args:
		n_classes (int): Number of output classes. If 0, the input is returned.
		If -1, all stages of the head, other than the final linear layer,
		are applied and the output returned. 
		Default is 0.
		pool_fn (T.Callable): Pooling function.
		Default is global_avg_pool.
		layer_norm_eps (T.Optional[float]): Epsilon value
		passed to layer normalization. If None, no normalization is applied,
		and norm_first is ignored.
		Default is None.
		norm_first (bool): Whether to apply layer normalization before
		pooling instead of after.
		Default is False.
		bias (bool): Whether the linear layer should have
		a bias term.
		Default is True.
	"""
	n_classes: int = 0
	pool_fn: T.Callable = global_avg_pool
	layer_norm_eps: T.Optional[float] = None
	norm_first: bool = False
	bias: bool = True

	@nn.compact
	def __call__(self, input):
		if self.n_classes == 0:
			return input

		if self.norm_first:
			output = nn.LayerNorm(
				epsilon=self.layer_norm_eps,
				)(input) if self.layer_norm_eps else input
			output = self.pool_fn(output, retain_spatial=False)
		
		else:
			output = self.pool_fn(input, retain_spatial=False)
			output = nn.LayerNorm(
				epsilon=self.layer_norm_eps,
				)(output) if self.layer_norm_eps else output
		
		if self.n_classes != -1:
			output = nn.Dense(
				features=self.n_classes,
				use_bias=self.bias,
				)(output)

		return output


class ViTHead(nn.Module):
	"""
	Head for ViTs with optional pooling of tokens, layer
	normalization, and a linear layer.

	Args:
		n_classes (int): Number of output classes. If 0, the input is returned.
		If -1, all stages of the head, other than the final linear layer,
		are applied and the output returned. 
		Default is 0.
		pool (bool): Whether to average pool the tokens for
		generating predictions. If False, the first token
		in the input, assumed to be the class token, is used
		to generate predictions.
		layer_norm_eps (T.Optional[float]): Epsilon value
		passed to layer normalization applied at the beginning.
		If None, no normalization is applied.
		Default is None.
		norm_first (bool): Whether to apply layer normalization before
		pooling/class token extraction instead of after.
		Default is True.
		bias (bool): Whether the linear layer should have
		a bias term.
		Default is True.
	"""
	n_classes: int = 0
	pool: bool = False
	layer_norm_eps: T.Optional[float] = None
	norm_first: bool = True
	bias: bool = True

	@nn.compact
	def __call__(self, input):
		if self.n_classes == 0:
			return input

		if self.norm_first:
			output = nn.LayerNorm(
				epsilon=self.layer_norm_eps,
				)(input) if self.layer_norm_eps else input
			output = jnp.mean(output, axis=-2) if self.pool else output[:, 0]
	
		else:
			output = jnp.mean(input, axis=-2) if self.pool else input[:, 0]
			output = nn.LayerNorm(
				epsilon=self.layer_norm_eps,
				)(output) if self.layer_norm_eps else output

		if self.n_classes != -1:
			output = nn.Dense(
				features=self.n_classes,
				use_bias=self.bias,
				)(output)

		return output



"""
Convolution similar to that of Flax but can automatically calulate
padding, accepts integer kernel size, etc.
"""


__all__ = [
	'Conv',
	]


import typing as T

from flax import linen as nn

from .tuplify import tuplify


def get_kernel_size_stride_padding(
	kernel_size: T.Union[T.Tuple[int, int], int] = 3,
	stride: T.Union[T.Tuple[int, int], int] = 1,
	padding: T.Optional[T.Union[str, int]] = None,
	dilation: int = 1,
	) -> T.Tuple[T.Tuple[int, int], T.Tuple[int, int], T.Tuple[T.Tuple[int, int], T.Tuple[int, int]]]:
	"""
	Gets the appropriate kernel size, stride, and padding to be passed to a convolution.

	Args:
		kernel_size (T.Union[T.Tuple[int, int], int]): Kernel size.
		If an int, this value is used along both spatial dimensions.
		Default is 3.
		stride (T.Union[T.Tuple[int, int], int]): Stride. If an int,
		this value is used along both spatial dimensions.
		Default is 1.
		padding (T.Optional[T.Union[str, int]]): Padding. If None,
		it is set so the spatial dimensions are exactly divided by stride.
		If an int, this value is used along both spatial dimensions.
		Default is None.
		dilation (int): Dilation.
		Default is 1.
	
	Returns (T.Tuple[T.Tuple[int, int], T.Tuple[int, int], T.Tuple[T.Tuple[int, int], T.Tuple[int, int]]]):
	Kernel size, stride, and padding.
	"""
	kernel_size = tuplify(kernel_size)
	stride = tuplify(stride)

	if isinstance(padding, int):
		padding = 2*(tuplify(padding),)

	elif padding is None:
		padding = (tuplify((dilation * (kernel_size[0]-1)) // 2), tuplify((dilation * (kernel_size[1]-1)) // 2))
	
	return kernel_size, stride, padding


class Conv(nn.Module):
	"""
	Similar to Flax's convolution but accepts integer kernel size, 
	supports depthwise convolution, etc.

	Args:
		out_dim (T.Optional[int]): Number of output channels.
		If None, it is set to the number of input channels.
		Default is None.
		kernel_size (T.Union[T.Tuple[int, int], int]): Kernel size.
		If an int, this value is used along both spatial dimensions.
		Default is 3.
		stride (T.Union[T.Tuple[int, int], int]): Stride. If an int,
		this value is used along both spatial dimensions.
		Default is 1.
		padding (T.Optional[T.Union[str, int]]): Padding. If None,
		it is set so the spatial dimensions are exactly divided by stride.
		If an int, this value is used along both spatial dimensions.
		Default is None.
		groups (T.Union[int, str]): Number of groups. If 'dw', a depthwise convolution is performed.
		Default is 1.
		dilation (int): Dilation.
		Default is 1.
		bias (bool): Whether to have a bias term.
		Default is True.
	"""
	out_dim: T.Optional[int] = None
	kernel_size: T.Union[T.Tuple[int, int], int] = 3
	stride: T.Union[T.Tuple[int, int], int] = 1
	padding: T.Optional[T.Union[str, int]] = None
	groups: T.Union[int, str] = 1
	dilation: int = 1
	bias: bool = True

	@nn.compact
	def __call__(self, input):
		in_dim = input.shape[-1]
		out_dim = self.out_dim or in_dim
		kernel_size, stride, padding = get_kernel_size_stride_padding(
			kernel_size=self.kernel_size,
			stride=self.stride,
			padding=self.padding,
			dilation=self.dilation,
			)
		return nn.Conv(
			features=out_dim,
			kernel_size=kernel_size,
			strides=stride,
			padding=padding,
			feature_group_count=in_dim if self.groups == 'dw' else self.groups,
			kernel_dilation=self.dilation,
			use_bias=self.bias,
			)(input)



"""
ResNet.
"""


import typing as T
from functools import partial

from flax import linen as nn

from .. import layers
from .factory import register_configs


class ResNetStem(nn.Module):
	"""
	ResNet stem.
	
	Args:
		out_dim (int): Number of output channels.
		Default is 64.
	"""
	out_dim: int = 64

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = layers.ConvBNAct(
			out_dim=self.out_dim,
			kernel_size=7,
			stride=2,
			act=nn.relu,
			)(input, training=training)
		output = layers.max_pool(
			input=output,
			stride=2,
			)
		return output
	

class ResNetDStem(nn.Module):
	"""
	ResNet-D stem.

	Args:
		out_dims (T.Tuple[int, int, int]): Number of output
		channels of each 3 x 3 convolution. 
		Default is (32, 32, 64).
		max_pool (bool): Whether to use max pool as the
		final downsampling module. If False, a 3 x 3 convolution
		is used.
		Default is True.
	"""
	out_dims: T.Tuple[int, int, int] = (32, 32, 64)
	max_pool: bool = True

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = layers.ConvBNAct(
			out_dim=self.out_dims[0],
			stride=2,
			act=nn.relu,
			)(input, training=training)
		output = layers.ConvBNAct(
			out_dim=self.out_dims[1],
			act=nn.relu,
			)(output, training=training)
		output = layers.ConvBNAct(
			out_dim=self.out_dims[2],
			act=nn.relu,
			)(output, training=training)

		if self.max_pool:
			output = layers.max_pool(
				input=output,
				stride=2,
				)
		
		else:
			output = layers.ConvBNAct(
				out_dim=self.out_dims[-1],
				stride=2,
				act=nn.relu,
				)(output, training=training)

		return output


class ResNetDownsample(nn.Module):
	"""
	ResNet downsampling module.

	Args:
		out_dim (int): Number of output channels.
		stride (int): Stride, i.e., spatial downsampling
		factor.
		Default is 1.
		avg_pool (bool): Whether to average pool for downsampling
		instead of a strided convolution.
		Default is False.
	"""
	out_dim: int
	stride: int = 1
	avg_pool: bool = False

	@nn.compact
	def __call__(self, input, training: bool = True):
		if self.stride != 1 and self.avg_pool:
			input = layers.avg_pool(
				input=input,
				kernel_size=self.stride,
				stride=self.stride,
				padding=0,
				)
		output = layers.ConvBNAct(
			out_dim=self.out_dim,
			kernel_size=1,
			stride=1 if self.avg_pool else self.stride,
			)(input, training=training)
		return output


class BasicBlock(nn.Module):
	"""
	Residual basic block.

	Args:
		bottleneck_dim (int): Number of output channels of the first convolution.
		out_dim (T.Optional[int]): Number of output channels of the second convolution.
		If None, it is set to bottleneck_dim.
		Default is None.
		conv_bn_relu (T.Callable): Layer used as the 
		first convolution-batch normalization-ReLU module.
		Default is layers.ConvBNAct.
		cardinality (int): Cardinality.
		Default is 1.
		stride (int): Stride.
		Default is 1.
		downsample (T.Callable): Downsampling module to use if input
		and output shapes are different.
		Default is ResNetDownsample.
		attention (T.Callable): T.Callable returning an attention module
		executed immediately after the second batch normalization.
		Default is Identity.
		avg_downsample (T.Optional[str]): If 'pre', average pooling is used
		before the first 3 x 3 convolution for spatial downsampling if stride is not 1,
		and the convolution's stride is set to 1. If 'post', average pooling is used
		after the first 3 x 3 convolution for spatial downsampling if stride is not 1,
		and the convolution's stride is set to 1. If None, spatial downsampling
		is performed by the first 3 x 3 convolution.
		Default is None.
	"""
	bottleneck_dim: int
	out_dim: T.Optional[int] = None
	conv_bn_relu: T.Callable = layers.ConvBNAct
	cardinality: int = 1
	stride: int = 1
	downsample: T.Callable = ResNetDownsample
	attention: T.Callable = layers.Identity
	avg_downsample: T.Optional[str] = None

	@nn.compact
	def __call__(self, input, training: bool = True):
		if self.avg_downsample == 'pre' and self.stride != 1:
			input = layers.avg_pool(
				input=input,
				stride=self.stride,
				)
		output = self.conv_bn_relu(
			out_dim=self.bottleneck_dim,
			stride=1 if self.avg_downsample else self.stride,
			groups=self.cardinality,
			act=nn.relu,
			)(input, training=training)
		if self.avg_downsample == 'post' and self.stride != 1:
			output = layers.avg_pool(
				input=output,
				stride=self.stride,
				)

		output = layers.ConvBNAct(
			out_dim=self.out_dim,
			groups=self.cardinality,
			)(output, training=training)
		output = self.attention()(output)

		if input.shape != output.shape:
			input = self.downsample(
				out_dim=self.out_dim or self.bottleneck_dim,
				stride=self.stride,
				)(input, training=training)
		
		output = input+output
		output = nn.relu(output)
		return output


class BottleneckBlock(nn.Module):
	"""
	Residual bottleneck block.

	Args:
		out_dim (int): Number of output channels.
		bottleneck_dim (int): Number of bottleneck channels.
		conv_bn_relu (T.Callable): Layer used as the 
		middle convolution-batch normalization-ReLU module.
		Default is layers.ConvBNAct.
		cardinality (int): Cardinality.
		Default is 1.
		stride (int): Stride.
		Default is 1.
		downsample (T.Callable): Downsampling module to use if input
		and output shapes are different.
		Default is ResNetDownsample.
		attention (T.Callable): T.Callable returning an attention module
		executed immediately after the third batch normalization.
		Default is Identity.
		avg_downsample (T.Optional[str]): If 'pre', average pooling is used
		before the 3 x 3 convolution for spatial downsampling if stride is not 1,
		and the convolution's stride is set to 1. If 'post', average pooling is used
		after the 3 x 3 convolution for spatial downsampling if stride is not 1,
		and the convolution's stride is set to 1. If None, spatial downsampling
		is performed by the 3 x 3 convolution.
		Default is None.
	"""
	out_dim: int
	bottleneck_dim: int
	conv_bn_relu: T.Callable = layers.ConvBNAct
	cardinality: int = 1
	stride: int = 1
	downsample: T.Callable = ResNetDownsample
	attention: T.Callable = layers.Identity
	avg_downsample: T.Optional[str] = None

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = layers.ConvBNAct(
			out_dim=self.bottleneck_dim,
			kernel_size=1,
			act=nn.relu,
			)(input, training=training)
		
		if self.avg_downsample == 'pre' and self.stride != 1:
			output = layers.avg_pool(
				input=output,
				stride=self.stride,
				)
		output = self.conv_bn_relu(
			stride=1 if self.avg_downsample else self.stride,
			groups=self.cardinality,
			act=nn.relu,
			)(output, training=training)
		if self.avg_downsample == 'post' and self.stride != 1:
			output = layers.avg_pool(
				input=output,
				stride=self.stride,
				)
		
		output = layers.ConvBNAct(
			out_dim=self.out_dim,
			kernel_size=1,
			)(output, training=training)
		output = self.attention()(output)
		
		if input.shape != output.shape:
			input = self.downsample(
				out_dim=self.out_dim,
				stride=self.stride,
				)(input, training=training)
		
		output = input+output
		output = nn.relu(output)
		
		return output


class ResNetStage(nn.Module):
	"""
	ResNet stage.

	Args:
		depth (int): Depth.
		out_dim (int): Number of output channels.
		bottleneck_dim (int): Number of bottleneck channels.
		block (T.Callable): Block used to construct the stage.
		Default is BottleneckBlock.
		conv_bn_relu (T.Callable): Layer used as the 
		first convolution-batch normalization-ReLU module
		of basic blocks or the middle convolution-batch normalization-ReLU
		module of bottleneck blocks.
		Default is layers.ConvBNAct.
		cardinality (int): Cardinality.
		Default is 1.
		stride (int): Stride.
		Default is 1.
		downsample (T.Callable): Downsampling module to use if input and output
		shapes are different.
		Default is ResNetDownsample.
		attention (T.Callable): T.Callable returning an attention module
		executed immediately after the third batch normalization in each block.
		Default is Identity.
		avg_downsample (T.Optional[str]): If 'pre', average pooling is used
		before the first 3 x 3 convolution of each block for spatial downsampling if stride is not 1,
		and the convolution's stride is set to 1. If 'post', average pooling is used
		after the first 3 x 3 convolution of each block for spatial downsampling if stride is not 1,
		and the convolution's stride is set to 1. If None, spatial downsampling
		is performed by the 3 x 3 convolutions of each block.
		Default is None.
	"""
	depth: int
	out_dim: int
	bottleneck_dim: int
	block: T.Callable = BottleneckBlock
	conv_bn_relu: T.Callable = layers.ConvBNAct
	cardinality: int = 1
	stride: int = 1
	downsample: T.Callable = ResNetDownsample
	attention: T.Callable = layers.Identity
	avg_downsample: T.Optional[str] = None

	@nn.compact
	def __call__(self, input, training: bool = True):
		for block_ind in range(self.depth):
			input = self.block(
				out_dim=self.out_dim,
				bottleneck_dim=self.bottleneck_dim,
				conv_bn_relu=self.conv_bn_relu,
				cardinality=self.cardinality,
				stride=self.stride if block_ind == 0 else 1,
				downsample=self.downsample,
				attention=self.attention,
				avg_downsample=self.avg_downsample,
				)(input, training=training)
		return input


class ResNet(nn.Module):
	"""
	ResNet, with support for different bottleneck dimensions, stems,
	etc.

	Args:
		depths (T.Tuple[int, ...]): Depth of each stage.
		block (T.Callable): Block used to construct the network.
		Default is BottleneckBlock.
		conv_bn_relu (T.Callable): Layer used as the 
		first convolution-batch normalization-ReLU module
		of basic blocks or the middle convolution-batch normalization-ReLU
		module of bottleneck blocks.
		Default is layers.ConvBNAct.
		cardinality (int): Cardinality.
		Default is 1.
		dim_per_cardinal_first_stage (int): Number of channels per cardinal group for the
		first stage.
		Default is 64.
		stem (T.Callable): Stem.
		Default is ResNetStem.
		downsample (T.Callable): Downsampling module to use if input and output
		shapes are different.
		Default is ResNetDownsample.
		attention (T.Callable): T.Callable returning an attention module executed
		immediately after the second batch normalization in each basic block or
		the third batch normalization in each bottleneck block.
		Default is Identity.
		avg_downsample (T.Optional[str]): If 'pre', average pooling is used
		before the first 3 x 3 convolution of each block for spatial downsampling if stride is not 1,
		and the convolution's stride is set to 1. If 'post', average pooling is used
		after the first 3 x 3 convolution of each block for spatial downsampling if stride is not 1,
		and the convolution's stride is set to 1. If None, spatial downsampling
		is performed by the first 3 x 3 convolution of each block.
		Default is None.
		n_classes (int): Number of output classes. If 0, there is no head,
		and the raw final features are returned. If -1, all stages of the 
		head, other than the final linear layer, are applied and the output
		returned.
		Default is 0.
	"""
	depths: T.Tuple[int, ...]
	block: T.Callable = BottleneckBlock
	conv_bn_relu: T.Callable = layers.ConvBNAct
	cardinality: int = 1
	dim_per_cardinal_first_stage: int = 64
	stem: T.Callable = ResNetStem
	downsample: T.Callable = ResNetDownsample
	attention: T.Callable = layers.Identity
	avg_downsample: T.Optional[str] = None
	n_classes: int = 0

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = self.stem()(input, training=training)
		self.sow(
			col='intermediates',
			name='stage_0',
			value=output,
			)

		for stage_ind, depth in enumerate(self.depths):
			bottleneck_dim = self.cardinality * self.dim_per_cardinal_first_stage * (2**stage_ind)
			output = ResNetStage(
				depth=depth,
				out_dim=bottleneck_dim if self.block == BasicBlock else 2 ** (stage_ind+8),
				bottleneck_dim=bottleneck_dim,
				block=self.block,
				conv_bn_relu=self.conv_bn_relu,
				cardinality=self.cardinality,
				stride=1 if stage_ind == 0 else 2,
				downsample=self.downsample,
				attention=self.attention,
				avg_downsample=self.avg_downsample,
				)(output, training=training)
			self.sow(
				col='intermediates',
				name=f'stage_{stage_ind+1}',
				value=output,
				)
		
		output = layers.Head(
			n_classes=self.n_classes,
			)(output)
		
		return output


@register_configs
def get_resnet_configs() -> T.Tuple[T.Type[ResNet], T.Dict]:
	"""
	Gets configurations for all available
	ResNet models.

	Returns (T.Tuple[T.Type[ResNet], T.Dict]): The ResNet class and
	configurations of all available models.
	"""
	configs = {
		'resnet18': {
			'depths': (2, 2, 2, 2),
			'block': BasicBlock,
			},
		'resnet34': {
			'depths': (3, 4, 6, 3),
			'block': BasicBlock,
			},
		'resnet26': {
			'depths': (2, 2, 2, 2),
			},
		'resnet50': {
			'depths': (3, 4, 6, 3),
			},
		'resnet101': {
			'depths': (3, 4, 23, 3),
			},
		'resnet152': {
			'depths': (3, 8, 36, 3),
			},
		'resnet18d': {
			'depths': (2, 2, 2, 2),
			'block': BasicBlock,
			'stem': ResNetDStem,
			'downsample': partial(ResNetDownsample, avg_pool=True),
			},
		'resnet34d': {
			'depths': (3, 4, 6, 3),
			'block': BasicBlock,
			'stem': ResNetDStem,
			'downsample': partial(ResNetDownsample, avg_pool=True),
			},
		'resnet26d': {
			'depths': (2, 2, 2, 2),
			'stem': ResNetDStem,
			'downsample': partial(ResNetDownsample, avg_pool=True),
			},
		'resnet50d': {
			'depths': (3, 4, 6, 3),
			'stem': ResNetDStem,
			'downsample': partial(ResNetDownsample, avg_pool=True),
			},
		'resnet101d': {
			'depths': (3, 4, 23, 3),
			'stem': ResNetDStem,
			'downsample': partial(ResNetDownsample, avg_pool=True),
			},
		'resnet152d': {
			'depths': (3, 8, 36, 3),
			'stem': ResNetDStem,
			'downsample': partial(ResNetDownsample, avg_pool=True),
			},
		'resnet200d': {
			'depths': (3, 24, 36, 3),
			'stem': ResNetDStem,
			'downsample': partial(ResNetDownsample, avg_pool=True),
			},
		'resnet10t': {
			'depths': (1, 1, 1, 1),
			'block': BasicBlock,
			'stem': partial(ResNetDStem, out_dims=(24, 32, 64)),
			'downsample': partial(ResNetDownsample, avg_pool=True),
			},
		'resnet14t': {
			'depths': (1, 1, 1, 1),
			'stem': partial(ResNetDStem, out_dims=(24, 32, 64)),
			'downsample': partial(ResNetDownsample, avg_pool=True),
			},
		'resnet26t': {
			'depths': (2, 2, 2, 2),
			'stem': partial(ResNetDStem, out_dims=(24, 32, 64)),
			'downsample': partial(ResNetDownsample, avg_pool=True),
			},
		'wide_resnet50_2': {
			'depths': (3, 4, 6, 3),
			'dim_per_cardinal_first_stage': 128,
			},
		'wide_resnet101_2': {
			'depths': (3, 4, 23, 3),
			'dim_per_cardinal_first_stage': 128,
			},
		'resnext50_32x4d': {
			'depths': (3, 4, 6, 3),
			'cardinality': 32,
			'dim_per_cardinal_first_stage': 4,
			},
		'resnext101_32x8d': {
			'depths': (3, 4, 23, 3),
			'cardinality': 32,
			'dim_per_cardinal_first_stage': 8,
			},
		'resnext101_64x4d': {
			'depths': (3, 4, 23, 3),
			'cardinality': 64,
			'dim_per_cardinal_first_stage': 4,
			},
		'seresnext50_32x4d': {
			'depths': (3, 4, 6, 3),
			'cardinality': 32,
			'dim_per_cardinal_first_stage': 4,
			'attention': layers.SE,
			},
		'seresnext101_32x8d': {
			'depths': (3, 4, 23, 3),
			'cardinality': 32,
			'dim_per_cardinal_first_stage': 8,
			'attention': layers.SE,
			},
		'seresnext26d_32x4d': {
			'depths': (2, 2, 2, 2),
			'cardinality': 32,
			'dim_per_cardinal_first_stage': 4,
			'stem': ResNetDStem,
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'attention': layers.SE,
			},
		'seresnext26t_32x4d': {
			'depths': (2, 2, 2, 2),
			'cardinality': 32,
			'dim_per_cardinal_first_stage': 4,
			'stem': partial(ResNetDStem, out_dims=(24, 32, 64)),
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'attention': layers.SE,
			},
		'ecaresnet50d': {
			'depths': (3, 4, 6, 3),
			'stem': ResNetDStem,
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'attention': layers.ECA,
			},
		'ecaresnet101d': {
			'depths': (3, 4, 23, 3),
			'stem': ResNetDStem,
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'attention': layers.ECA,
			},
		'ecaresnet269d': {
			'depths': (3, 30, 48, 8),
			'stem': ResNetDStem,
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'attention': layers.ECA,
			},
		'ecaresnet26t': {
			'depths': (2, 2, 2, 2),
			'stem': partial(ResNetDStem, out_dims=(24, 32, 64)),
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'attention': layers.ECA,
			},
		'ecaresnet50t': {
			'depths': (3, 4, 6, 3),
			'stem': partial(ResNetDStem, out_dims=(24, 32, 64)),
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'attention': layers.ECA,
			},
		'resnetrs50': {
			'depths': (3, 4, 6, 3),
			'stem': partial(ResNetDStem, max_pool=False),
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'attention': partial(layers.SE, reduction_factor=4),
			},
		'resnetrs101': {
			'depths': (3, 4, 23, 3),
			'stem': partial(ResNetDStem, max_pool=False),
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'attention': partial(layers.SE, reduction_factor=4),
			},
		'resnetrs152': {
			'depths': (3, 8, 36, 3),
			'stem': partial(ResNetDStem, max_pool=False),
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'attention': partial(layers.SE, reduction_factor=4),
			},
		'resnetrs200': {
			'depths': (3, 24, 36, 3),
			'stem': partial(ResNetDStem, max_pool=False),
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'attention': partial(layers.SE, reduction_factor=4),
			},
		'resnetrs270': {
			'depths': (4, 29, 53, 4),
			'stem': partial(ResNetDStem, max_pool=False),
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'attention': partial(layers.SE, reduction_factor=4),
			},
		'resnetrs350': {
			'depths': (4, 36, 72, 4),
			'stem': partial(ResNetDStem, max_pool=False),
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'attention': partial(layers.SE, reduction_factor=4),
			},
		'resnetrs420': {
			'depths': (4, 44, 87, 4),
			'stem': partial(ResNetDStem, max_pool=False),
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'attention': partial(layers.SE, reduction_factor=4),
			},
		'skresnet18': {
			'depths': (2, 2, 2, 2),
			'block': BasicBlock,
			'conv_bn_relu': partial(layers.SK),
			},
		'skresnet34': {
			'depths': (3, 4, 6, 3),
			'block': BasicBlock,
			'conv_bn_relu': partial(layers.SK),
			},
		'skresnext50_32x4d': {
			'depths': (3, 4, 6, 3),
			'conv_bn_relu': partial(layers.SK, reduction_factor=16,	min_reduction_dim=32, split=False),
			'cardinality': 32,
			'dim_per_cardinal_first_stage': 4,
			},
		'resnest14_2s1x64d': {
			'depths': (1, 1, 1, 1),
			'conv_bn_relu': layers.SplAt,
			'stem': ResNetDStem,
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'avg_downsample': 'post',
			},
		'resnest26_2s1x64d': {
			'depths': (2, 2, 2, 2),
			'conv_bn_relu': layers.SplAt,
			'stem': ResNetDStem,
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'avg_downsample': 'post',
			},
		'resnest50_2s1x64d': {
			'depths': (3, 4, 6, 3),
			'conv_bn_relu': layers.SplAt,
			'stem': ResNetDStem,
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'avg_downsample': 'post',
			},
		'resnest101_2s1x64d': {
			'depths': (3, 4, 23, 3),
			'conv_bn_relu': layers.SplAt,
			'stem': partial(ResNetDStem, out_dims=(64, 64, 128)),
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'avg_downsample': 'post',
			},
		'resnest200_2s1x64d': {
			'depths': (3, 24, 36, 3),
			'conv_bn_relu': layers.SplAt,
			'stem': partial(ResNetDStem, out_dims=(64, 64, 128)),
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'avg_downsample': 'post',
			},
		'resnest269_2s1x64d': {
			'depths': (3, 30, 48, 8),
			'conv_bn_relu': layers.SplAt,
			'stem': partial(ResNetDStem, out_dims=(64, 64, 128)),
			'downsample': partial(ResNetDownsample, avg_pool=True),
			'avg_downsample': 'post',
			},
		}
	return ResNet, configs



"""
Large kernel attention (LKA) by Guo et al.

References:
- Guo et al. Visual Attention Network
"""


__all__ = [
	'LKA',
	]


from flax import linen as nn

from .conv import Conv


class LKA(nn.Module):
	"""
	Large kernel attention.

	Args:
		dw_kernel_size (int): Kernel size of
		the depthwise convolution.
		Default is 5.
		dilated_dw_kernel_size (int): Kernel size of
		the dilated depthwise convolution.
		Default is 7.
	"""
	dw_kernel_size: int = 5
	dilated_dw_kernel_size: int = 7

	@nn.compact
	def __call__(self, input):
		attention = Conv(
			kernel_size=self.dw_kernel_size,
			groups='dw',
			)(input)
		attention = Conv(
			kernel_size=self.dilated_dw_kernel_size,
			groups='dw',
			dilation=self.dilated_dw_kernel_size//2,
			)(attention)
		attention = Conv(
			kernel_size=1,
			)(attention)
		return input*attention



"""
Pyramid vision transformer V2 (PVT V2).
"""


import typing as T
from functools import partial
from math import sqrt

from flax import linen as nn
from jax import numpy as jnp

from .. import layers
from .factory import register_configs


class SRAQKV(nn.Module):
	"""
	Query-key-value extractor for spatial reduction attention. 

	Args:
		n_heads (int): Number of heads.
		sr_factor (int): Reduction factor for spatial
		downsampling of the key and value vectors.
		Default is 1.
	"""
	n_heads: int
	sr_factor: int = 1

	@nn.compact
	def __call__(self, input):
		input_kv = input
		token_dim = input.shape[-1]
		head_dim = token_dim//self.n_heads
		
		if 1 < self.sr_factor:
			img_size = int(sqrt(input.shape[-2]))
			input_kv = jnp.reshape(input, (-1, img_size, img_size, token_dim))
			input_kv = layers.PatchEmbed(
				token_dim=token_dim,
				patch_size=self.sr_factor,
				layer_norm_eps=1e-5,
				)(input_kv)
		
			q = nn.Dense(
				features=token_dim,
				)(input)
			kv = nn.Dense(
				features=2*token_dim,
				)(input_kv)
		
			q = jnp.reshape(q, (len(input), -1, self.n_heads, head_dim))
			kv = jnp.reshape(kv, (len(input), -1, 2, self.n_heads, head_dim))
		
			q = jnp.swapaxes(q, axis1=1, axis2=2)
			kv = jnp.transpose(kv, (2, 0, 3, 1, 4))
			k, v = jnp.split(
				ary=kv,
				indices_or_sections=2,
				axis=0,
				)
			k, v = jnp.squeeze(k, axis=0), jnp.squeeze(v, axis=0)
		
		else:
			q, k, v = layers.QKV(
				n_heads=self.n_heads,
				)(input)
		
		return q, k, v
		

class PVTV2Stage(nn.Module):
	"""
	PVTV2 stage.

	Args:
		depths (int): Depth.
		token_dim (int): Token dimension.
		n_heads (int): Number of heads.
		sr_factor (int): Reduction factor for spatial
		downsampling of the key and value vectors.
		Default is 1.
		mlp_hidden_dim_expansion_factor (float): Factor of expansion
		for the hidden neurons of the MLP.
		Default is 4.
		downsample (bool): Whether to downsample.
		Default is False.
	"""
	depth: int
	token_dim: int
	n_heads: int
	sr_factor: int = 1
	mlp_hidden_dim_expansion_factor: float = 4.
	downsample: bool = False

	@nn.compact
	def __call__(self, input):
		if self.downsample:
			input = layers.PatchEmbed(
				token_dim=self.token_dim,
				patch_size=3,
				patch_stride=2,
				layer_norm_eps=1e-5,
				)(input)

		for _ in range(self.depth):
			input = layers.MetaFormerBlock(
				token_mixer=lambda: layers.MHSA(
					to_qkv=partial(
						SRAQKV,
						n_heads=self.n_heads,
						sr_factor=self.sr_factor,
						),
					),
				mlp_hidden_dim_expansion_factor=self.mlp_hidden_dim_expansion_factor,
				dw_kernel_size=3,
				)(input)
		output = nn.LayerNorm()(input)

		img_size = int(sqrt(output.shape[-2]))
		output = jnp.reshape(output, (len(input), img_size, img_size, -1))
		return output


class PVTV2(nn.Module):
	"""
	Pyramid vision transformer V2.

	Args:
		depths (T.Tuple[int, ...]): Depth of each stage.
		token_dims (T.Tuple[int, ...]): Token dimension of each stage.
		n_heads (T.Tuple[int, ...]): Number of heads of each
		stage.
		sr_factors (T.Tuple[int, ...]): Reduction factor for spatial
		reduction attention of each stage.
		Default is (8, 4, 2, 1).
		mlp_hidden_dim_expansion_factors (T.Tuple[float, ...]): Factor of expansion
		for the hidden neurons of the MLP of each stage.
		Default is (8, 8, 4, 4).
		n_classes (int): Number of output classes. If 0, there is no head,
		and the raw final features are returned. If -1, all stages of the 
		head, other than the final linear layer, are applied and the output
		returned.
		Default is 0.
	"""
	depths: T.Tuple[int, ...]
	token_dims: T.Tuple[int, ...]
	n_heads: T.Tuple[int, ...]
	sr_factors: T.Tuple[int, ...] = (8, 4, 2, 1)
	mlp_hidden_dim_expansion_factors: T.Tuple[float, ...] = (8., 8., 4., 4.)
	n_classes: int = 0

	@nn.compact
	def __call__(self, input):
		output = layers.PatchEmbed(
			token_dim=self.token_dims[0],
			patch_size=7,
			patch_stride=4,
			layer_norm_eps=1e-5,
			)(input)
		self.sow(
			col='intermediates',
			name='stage_0',
			value=output,
			)
		
		for stage_ind in range(len(self.depths)):
			output = PVTV2Stage(
				depth=self.depths[stage_ind],
				token_dim=self.token_dims[stage_ind],
				n_heads=self.n_heads[stage_ind],
				sr_factor=self.sr_factors[stage_ind],
				mlp_hidden_dim_expansion_factor=self.mlp_hidden_dim_expansion_factors[stage_ind],
				downsample=False if stage_ind == 0 else True,
				)(output)
			self.sow(
				col='intermediates',
				name=f'stage_{stage_ind+1}',
				value=output,
				)
		
		output = layers.Head(
			n_classes=self.n_classes,
			)(output)
		
		return output


@register_configs
def get_pvtv2_configs() -> T.Tuple[T.Type[PVTV2], T.Dict]:
	"""
	Gets configurations for all available
	PVTV2 models.

	Returns (T.Tuple[T.Type[PVTV2], T.Dict]): PVTV2 class and
	configurations of all models.
	"""
	configs = {
		'pvtv2_b0': {
			'depths': (2, 2, 2, 2),
			'token_dims': (32, 64, 160, 256),
			'n_heads': (1, 2, 5, 8),
			},
		'pvtv2_b1': {
			'depths': (2, 2, 2, 2),
			'token_dims': (64, 128, 320, 512),
			'n_heads': (1, 2, 5, 8),
			},
		'pvtv2_b2': {
			'depths': (3, 4, 6, 3),
			'token_dims': (64, 128, 320, 512),
			'n_heads': (1, 2, 5, 8),
			},
		'pvtv2_b3': {
			'depths': (3, 4, 18, 3),
			'token_dims': (64, 128, 320, 512),
			'n_heads': (1, 2, 5, 8),
			},
		'pvtv2_b4': {
			'depths': (3, 8, 27, 3),
			'token_dims': (64, 128, 320, 512),
			'n_heads': (1, 2, 5, 8),
			},
		'pvtv2_b5': {
			'depths': (3, 6, 40, 3),
			'token_dims': (64, 128, 320, 512),
			'n_heads': (1, 2, 5, 8),
			'mlp_hidden_dim_expansion_factors': (4, 4, 4, 4),
			},
		}
	return PVTV2, configs



"""
Relative position embedding by Huang et al., code borrowed from Wightman.

References: 
- Huang et al. Music Transformer: Generating Music with Long-Term Structure.
- Wightman. PyTorch Image Models.
"""


__all__ = [
	'RelPosEmbed',
	]


import typing as T

from flax import linen as nn
from jax import numpy as jnp

from .tuplify import tuplify


def get_n_rel_distance(
	window_size: T.Union[T.Tuple[int, int], int],
	class_token: bool = True,
	) -> int:
	"""
	Gets the number of relative distances for
	relative position embedding.

	Args:
		window_size (T.Union[T.Tuple[int, int], int]): Window size.
		If an int, this value is used along both spatial dimesnoins.
		class_token (bool): Whether the input contains a class token.
		Default is True.
	"""
	window_size = tuplify(window_size)
	n_rel_distance = (2*window_size[0] - 1) * (2*window_size[1] - 1)
	return n_rel_distance+3 if class_token else n_rel_distance


def get_rel_pos_ind(
	window_size: T.Union[T.Tuple[int, int], int],
	class_token: bool = True,
	) -> jnp.ndarray:
	"""
	Gets a matrix used to index a relative position bias
	table.

	Args:
		window_size (T.Union[T.Tuple[int, int], int]): Window size.
		If an int, this value is used along both spatial dimesnoins.
		class_token (bool): Whether the input contains a class token.
		Default is True.
	
	Returns (jnp.ndarray): Matrix used to index a relative position bias
	table.
	"""
	window_h, window_w = tuplify(window_size)

	coords_h = jnp.arange(window_h)
	coords_w = jnp.arange(window_w)
	coords = jnp.meshgrid(coords_h, coords_w, indexing='ij')
	coords = jnp.stack(coords)
	coords = coords.reshape(len(coords), -1)

	rel_coords = coords[:, :, None] - coords[:, None, :]
	rel_coords = jnp.transpose(rel_coords, (1, 2, 0))
	
	rel_coords = rel_coords.at[:, :, 0].set(rel_coords[:, :, 0] + window_h - 1)
	rel_coords = rel_coords.at[:, :, 1].set(rel_coords[:, :, 1] + window_w - 1)
	rel_coords = rel_coords.at[:, :, 0].set(rel_coords[:, :, 0] * (2 * window_w - 1))

	if not class_token:
		return jnp.sum(rel_coords, -1)

	n_rel_distance = get_n_rel_distance(window_h, window_w)
	area = window_h*window_w

	rel_pos_ind = jnp.zeros(tuplify(area+1), dtype=int)
	rel_pos_ind = rel_pos_ind.at[1:, 1:].set(jnp.sum(rel_coords, -1))
	rel_pos_ind = rel_pos_ind.at[0, 0:].set(n_rel_distance-3)
	rel_pos_ind = rel_pos_ind.at[0:, 0].set(n_rel_distance-2)
	rel_pos_ind = rel_pos_ind.at[0, 0].set(n_rel_distance-1)

	return rel_pos_ind


class RelPosEmbed(nn.Module):
	"""
	Relative position embedding.

	Args:
		n_heads (int): Number of heads.
		window_size (T.Union[T.Tuple[int, int], int]): Window size.
		If an int, this value is used along both spatial dimesnoins.
		class_token (bool): Whether the input contains a class token.
		Default is True.
	"""
	n_heads: int
	window_size: T.Union[T.Tuple[int, int], int]
	class_token: bool = True

	@nn.compact
	def __call__(self, input):
		n_rel_distance = get_n_rel_distance(
			window_size=self.window_size,
			class_token=self.class_token,
			)
		rel_pos_table = nn.Embed(
			num_embeddings=n_rel_distance,
			features=self.n_heads,
			)
		rel_pos_ind = self.variable(
			col='rel_pos_ind',
			name='rel_pos_ind',
			init_fn=lambda: get_rel_pos_ind(
				window_size=self.window_size,
				class_token=self.class_token,
				),
			).value

		rel_pos_bias = rel_pos_table(rel_pos_ind)
		rel_pos_bias = jnp.transpose(rel_pos_bias, (2, 0, 1))
		
		return input+rel_pos_bias



"""
Common deep learning modules in computer vision.
"""


from .acts import *
from .conv import *
from .eca import * 
from .gn_conv import *
from .head import *
from .identity import *
from .layer_scale import *
from .lin_norm_act import *
from .lka import *
from .meta_former import *
from .mhsa import *
from .mlp import *
from .pool import *
from .rel_pos_embed import *
from .se import *
from .sk import *
from .splat import *
from .tuplify import *
from .vit_layers import *
from .window import *



"""
Vision transformer (ViT).
"""


import typing as T
from functools import partial

from flax import linen as nn

from .. import layers
from .factory import register_configs


class ViT(nn.Module):
	"""
	Vision transformer.

	Args:
		depth (int): Depth.
		token_dim (int): Token dimension.
		n_heads (int): Number of heads.
		patch_size (int): Patch size. This value is used
		along both spatial dimensions.
		Default is 16.
		mlp_hidden_dim_expansion_factor (float): Factor of expansion
		for the hidden neurons of the MLP.
		Default is 4.
		layer_scale_init_value (T.Optional[float]): Value
		for initializing LayerScale. If None, no LayerScale
		is applied.
		Default is None.
		layer_norm_eps (float): Epsilon value passed to layer
		normalization.
		Default is 1e-6.
		k_bias (bool): Whether the linear transformation
		for obtaining keys should have a bias term.
		Default is True.
		rel_pos_embed (bool): Whether to use relative position
		embedding instead of absolute position embeddings. 
		If True, class_token_pos_embed is ignored.
		Default is False.
		pre_norm (bool): Whether to have layer normalization
		immediately before the body of the network.
		Default is False.
		class_token_pos_embed (bool): Whether to apply position 
		embedding to the class token.
		Default is True.
		n_classes (int): Number of output classes. If 0, there is no head,
		head_pool and head_norm_first are ignored, and the raw final features are returned. If -1,
		all stages of the head, other than the final linear layer, are applied
		and the output returned.
		Default is 0.
		head_pool (bool): Whether to average pool the tokens for generating
		predictions. If False, the class token is used to generate predictions.
		Default is False.
		head_norm_first (bool): Whether to apply layer normalization before
		pooling/class token extraction instead of after for the head.
		Default is True.
	"""
	depth: int
	token_dim: int
	n_heads: int
	patch_size: int = 16
	mlp_hidden_dim_expansion_factor: float = 4
	layer_scale_init_value: T.Optional[float] = None
	layer_norm_eps: float = 1e-6
	k_bias: bool = True
	rel_pos_embed: bool = False
	pre_norm: bool = False
	class_token_pos_embed: bool = True
	n_classes: int = 0
	head_pool: bool = False
	head_norm_first: bool = True

	@nn.compact
	def __call__(self, input):
		output = layers.PatchEmbed(
			token_dim=self.token_dim,
			patch_size=self.patch_size,
			bias=not self.pre_norm,
			)(input)
			
		if not self.rel_pos_embed:
			if self.class_token_pos_embed:
				output = layers.ClassToken()(output)
				output = layers.AbsPosEmbed()(output)
			
			else:
				output = layers.AbsPosEmbed()(output)
				output = layers.ClassToken()(output)
		
		else:
			output = layers.ClassToken()(output)
		
		output = nn.LayerNorm(
			epsilon=self.layer_norm_eps,
			)(output) if self.pre_norm else output
		self.sow(
			col='intermediates',
			name='block_0',
			value=output,
			)
		
		window_size = (input.shape[-3]//self.patch_size, input.shape[-2]//self.patch_size) if self.rel_pos_embed else None
		for block_ind in range(self.depth):
			output = layers.MetaFormerBlock(
				token_mixer=lambda: layers.MHSA(
					to_qkv=partial(
						layers.QKV,
						n_heads=self.n_heads,
						k_bias=self.k_bias,
						),
					pre_softmax=partial(
						layers.RelPosEmbed,
						n_heads=self.n_heads,
						window_size=window_size,
						) if window_size else layers.Identity,
					),
				mlp_hidden_dim_expansion_factor=self.mlp_hidden_dim_expansion_factor,
				layer_norm_eps=self.layer_norm_eps,
				layer_scale_init_value=self.layer_scale_init_value,
				)(output)
			self.sow(
				col='intermediates',
				name=f'block_{block_ind+1}',
				value=output,
				)

		if self.head_pool and self.n_classes != 0:
			output = output[:, 1:]
	
		output = layers.ViTHead(
			n_classes=self.n_classes,
			pool=self.head_pool,
			layer_norm_eps=self.layer_norm_eps,
			norm_first=self.head_norm_first,
			)(output)
		
		return output


@register_configs
def get_vit_configs() -> T.Tuple[T.Type[ViT], T.Dict]:
	"""
	Gets configurations for all available
	ViT models.

	Returns (T.Tuple[T.Type[ViT], T.Dict]): The ViT class and
	configurations of all available models.
	"""
	configs = {
		'vit_small_patch32_224': {
			'depth': 12,
			'token_dim': 384,
			'n_heads': 6,
			'patch_size': 32,
			},
		'vit_base_patch32_224': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			'patch_size': 32,
			},
		'vit_tiny_patch16_224': {
			'depth': 12,
			'token_dim': 192,
			'n_heads': 3,
			}, 
		'vit_small_patch16_224': {
			'depth': 12,
			'token_dim': 384,
			'n_heads': 6,
			}, 
		'vit_base_patch16_224': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			},
		'vit_large_patch16_224': {
			'depth': 24,
			'token_dim': 1024,
			'n_heads': 16,
			},
		'vit_base_patch8_224': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			'patch_size': 8,
			},
		'vit_small_patch32_384': {
			'depth': 12,
			'token_dim': 384,
			'n_heads': 6,
			'patch_size': 32,
			},
		'vit_base_patch32_384': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			'patch_size': 32,
			},
		'vit_large_patch32_384': {
			'depth': 24,
			'token_dim': 1024,
			'n_heads': 16,
			'patch_size': 32,
			},
		'vit_tiny_patch16_384': {
			'depth': 12,
			'token_dim': 192,
			'n_heads': 3,
			},
		'vit_small_patch16_384': {
			'depth': 12,
			'token_dim': 384,
			'n_heads': 6,
			},
		'vit_base_patch16_384': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			},
		'vit_large_patch16_384': {
			'depth': 24,
			'token_dim': 1024,
			'n_heads': 16,
			},
		'vit_small_patch32_224_in22k': {
			'depth': 12,
			'token_dim': 384,
			'n_heads': 6,
			'patch_size': 32,
			},
		'vit_base_patch32_224_in22k': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			'patch_size': 32,
			},
		'vit_large_patch32_224_in22k': {
			'depth': 24,
			'token_dim': 1024,
			'n_heads': 16,
			'patch_size': 32,
			},
		'vit_tiny_patch16_224_in22k': {
			'depth': 12,
			'token_dim': 192,
			'n_heads': 3,
			},
		'vit_small_patch16_224_in22k': {
			'depth': 12,
			'token_dim': 384,
			'n_heads': 6,
			},
		'vit_base_patch16_224_in22k': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			},
		'vit_large_patch16_224_in22k': {
			'depth': 24,
			'token_dim': 1024,
			'n_heads': 16,
			},
		'vit_huge_patch14_224_in22k': {
			'depth': 32,
			'token_dim': 1280,
			'n_heads': 16,
			'patch_size': 14,
			},
		'vit_base_patch8_224_in22k': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			'patch_size': 8,
			},
		'deit3_small_patch16_224': {
			'depth': 12,
			'token_dim': 384, 
			'n_heads': 6,
			'layer_scale_init_value': 1e-6,
			'class_token_pos_embed': False,
			},
		'deit3_medium_patch16_224': {
			'depth': 12,
			'token_dim': 512,
			'n_heads': 8,
			'layer_scale_init_value': 1e-6,
			'class_token_pos_embed': False,
			},
		'deit3_base_patch16_224': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			'layer_scale_init_value': 1e-6,
			'class_token_pos_embed': False,
			},
		'deit3_large_patch16_224': {
			'depth': 24,
			'token_dim': 1024,
			'n_heads': 16,
			'layer_scale_init_value': 1e-6,
			'class_token_pos_embed': False,
			},
		'deit3_huge_patch14_224': {
			'depth': 32,
			'token_dim': 1280,
			'n_heads': 16,
			'patch_size': 14,
			'layer_scale_init_value': 1e-6,
			'class_token_pos_embed': False,
			},
		'deit3_small_patch16_384': {
			'depth': 12,
			'token_dim': 384,
			'n_heads': 6,
			'layer_scale_init_value': 1e-6,
			'class_token_pos_embed': False,
			},
		'deit3_base_patch16_384': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			'layer_scale_init_value': 1e-6,
			'class_token_pos_embed': False,
			},
		'deit3_large_patch16_384': {
			'depth': 24,
			'token_dim': 1024,
			'n_heads': 16,
			'layer_scale_init_value': 1e-6,
			'class_token_pos_embed': False,
			},
		'deit3_small_patch16_224_in22ft1k': {
			'depth': 12,
			'token_dim': 384,
			'n_heads': 6,
			'layer_scale_init_value': 1e-6,
			'class_token_pos_embed': False,
			},
		'deit3_medium_patch16_224_in22ft1k': {
			'depth': 12,
			'token_dim': 512,
			'n_heads': 8,
			'layer_scale_init_value': 1e-6,
			'class_token_pos_embed': False,
			},
		'deit3_base_patch16_224_in22ft1k': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			'layer_scale_init_value': 1e-6,
			'class_token_pos_embed': False,
			},
		'deit3_large_patch16_224_in22ft1k': {
			'depth': 24,
			'token_dim': 1024,
			'n_heads': 16, 
			'layer_scale_init_value': 1e-6,
			'class_token_pos_embed': False,
			},
		'deit3_huge_patch14_224_in22ft1k': {
			'depth': 32,
			'token_dim': 1280,
			'n_heads': 16,
			'patch_size': 14,
			'layer_scale_init_value': 1e-6,
			'class_token_pos_embed': False,
			},
		'deit3_small_patch16_384_in22ft1k': {
			'depth': 12,
			'token_dim': 384,
			'n_heads': 6,
			'layer_scale_init_value': 1e-6,
			'class_token_pos_embed': False,
			},
		'deit3_base_patch16_384_in22ft1k': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			'layer_scale_init_value': 1e-6,
			'class_token_pos_embed': False,
			},
		'deit3_large_patch16_384_in22ft1k': {
			'depth': 24,
			'token_dim': 1024,
			'n_heads': 16,
			'layer_scale_init_value': 1e-6,
			'class_token_pos_embed': False,
			},
		'beit_base_patch16_224': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			'layer_scale_init_value': 1e-1,
			'rel_pos_embed': True,
			'k_bias': False,
			'head_pool': True,
			'head_norm_first': False,
			},
		'beit_large_patch16_224': {
			'depth': 24,
			'token_dim': 1024,
			'n_heads': 16,
			'layer_scale_init_value': 1e-5,
			'rel_pos_embed': True,
			'k_bias': False,
			'head_pool': True,
			'head_norm_first': False,
			},
		'beit_base_patch16_384': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			'layer_scale_init_value': 1e-1,
			'rel_pos_embed': True,
			'k_bias': False,
			'head_pool': True,
			'head_norm_first': False,
			},
		'beit_large_patch16_384': {
			'depth': 24,
			'token_dim': 1024,
			'n_heads': 16,
			'layer_scale_init_value': 1e-5,
			'rel_pos_embed': True,
			'k_bias': False,
			'head_pool': True,
			'head_norm_first': False,
			},
		'beit_large_patch16_512': {
			'depth': 24,
			'token_dim': 1024,
			'n_heads': 16,
			'layer_scale_init_value': 1e-5,
			'rel_pos_embed': True,
			'k_bias': False,
			'head_pool': True,
			'head_norm_first': False,
			},
		'beit_base_patch16_224_in22k': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			'layer_scale_init_value': 1e-1,
			'rel_pos_embed': True,
			'k_bias': False,
			'head_pool': True,
			'head_norm_first': False,
			},
		'beit_large_patch16_224_in22k': {
			'depth': 24,
			'token_dim': 1024,
			'n_heads': 16,
			'layer_scale_init_value': 1e-5,
			'rel_pos_embed': True,
			'k_bias': False,
			'head_pool': True,
			'head_norm_first': False,
			},
		'beitv2_base_patch16_224': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			'layer_scale_init_value': 1e-5,
			'rel_pos_embed': True,
			'k_bias': False,
			'head_pool': True,
			'head_norm_first': False,
			},
		'beitv2_large_patch16_224': {
			'depth': 24,
			'token_dim': 1024,
			'n_heads': 16,
			'layer_scale_init_value': 1e-5,
			'rel_pos_embed': True,
			'k_bias': False,
			'head_pool': True,
			'head_norm_first': False,
			},
		'beitv2_base_patch16_224_in22k': {
			'depth': 12,
			'token_dim': 768,
			'n_heads': 12,
			'layer_scale_init_value': 1e-5,
			'rel_pos_embed': True,
			'k_bias': False,
			'head_pool': True,
			'head_norm_first': False,
			},
		'beitv2_large_patch16_224_in22k': {
			'depth': 24,
			'token_dim': 1024,
			'n_heads': 16,
			'layer_scale_init_value': 1e-5,
			'rel_pos_embed': True,
			'k_bias': False,
			'head_pool': True,
			'head_norm_first': False,
			},
		}
	return ViT, configs



"""
ConvNeXt.
"""


import typing as T

from flax import linen as nn

from .. import layers
from .factory import register_configs


class ConvNeXtBlock(nn.Module):
	"""
	ConvNeXt block. 

	Args:
		out_dim (T.Optional[int]): Number of output channels.
		If None, it is set to the number of input channels.
		Default is None.
		layer_scale_init_value (T.Optional[float]): Value for initializing
		LayerScale. If None, no LayerScale is applied.
		Default is 1e-6.
	"""
	out_dim: T.Optional[int] = None
	layer_scale_init_value: T.Optional[float] = 1e-6

	@nn.compact
	def __call__(self, input):
		output = layers.Conv(
			out_dim=self.out_dim,
			kernel_size=7,
			groups='dw',
			)(input)
		output = layers.TransformerMLP(
			layer_scale_init_value=self.layer_scale_init_value,
			residual=False,
			)(output)
		return input+output


class ConvNeXtStage(nn.Module):
	"""
	ConvNeXt stage.

	Args:
		depth (int): Depth.
		out_dim (int): Number of output channels.
		stride (int): Stride.
		Default is 1.
		layer_scale_init_value (T.Optional[float]): Value for initializing
		LayerScale. If None, no LayerScale is applied.
		Default is 1e-6.
	"""
	depth: int
	out_dim: int
	stride: int = 1
	layer_scale_init_value: T.Optional[float] = 1e-6

	@nn.compact
	def __call__(self, input):
		if (self.stride != 1) or (input.shape[-1] != self.out_dim):
			input = layers.PatchEmbed(
				token_dim=self.out_dim,
				patch_size=self.stride,
				layer_norm_eps=1e-6,
				norm_first=True,
				flatten=False,
				)(input)

		for _ in range(self.depth):
			input = ConvNeXtBlock(
				layer_scale_init_value=self.layer_scale_init_value,
				)(input)

		return input


class ConvNeXt(nn.Module):
	"""
	ConvNeXt.

	Args:
		depths (T.Tuple[int, ...]): Depth of each stage.
		out_dims (T.Tuple[int, ...]): Number of output channels of each stage.
		layer_scale_init_value (T.Optional[float]): Value for initializing
		LayerScale. If None, no LayerScale is applied.
		Default is 1e-6.
		n_classes (int): Number of output classes. If 0, there is no head,
		and the raw final features are returned. If -1, all stages of the 
		head, other than the final linear layer, are applied and the output
		returned.
		Default is 0.
	"""
	depths: T.Tuple[int, ...]
	out_dims: T.Tuple[int, ...]
	layer_scale_init_value: T.Optional[float] = 1e-6
	n_classes: int = 0

	@nn.compact
	def __call__(self, input):
		output = layers.PatchEmbed(
			token_dim=self.out_dims[0],
			patch_size=4,
			layer_norm_eps=1e-6,
			flatten=False,
			)(input)
		self.sow(
			col='intermediates',
			name='stage_0',
			value=output,
			)

		for stage_ind in range(len(self.depths)):
			output = ConvNeXtStage(
				depth=self.depths[stage_ind],
				out_dim=self.out_dims[stage_ind],
				stride=1 if stage_ind == 0 else 2,
				layer_scale_init_value=self.layer_scale_init_value,
				)(output)
			self.sow(
				col='intermediates',
				name=f'stage_{stage_ind+1}',
				value=output,
				)
		
		output = layers.Head(
			n_classes=self.n_classes,
			layer_norm_eps=1e-6,
			)(output)
			
		return output


@register_configs
def get_convnext_configs() -> T.Tuple[T.Type[ConvNeXt], T.Dict]:
	"""
	Gets configurations for all available
	ConvNeXt models.

	Returns (T.Tuple[T.Type[ConvNeXt], T.Dict]): The ConvNeXt class and
	configurations of all available models.
	"""
	configs = {
		'convnext_xxxnano': {
			'depths': (2, 2, 6, 2),
			'out_dims': (40, 80, 160, 320),
			},
		'convnext_xxnano': {
			'depths': (2, 2, 6, 2),
			'out_dims': (48, 96, 192, 384),
			},
		'convnext_xnano': {
			'depths': (2, 2, 6, 2),
			'out_dims': (64, 128, 256, 512),
			},
		'convnext_nano': {
			'depths': (2, 2, 8, 2),
			'out_dims': (80, 160, 320, 640),
			},
		'convnext_tiny': {
			'depths': (3, 3, 9, 3),
			'out_dims': (96, 192, 384, 768),
			},
		'convnext_small': {
			'depths': (3, 3, 27, 3),
			'out_dims': (96, 192, 384, 768),
			},
		'convnext_base': {
			'depths': (3, 3, 27, 3),
			'out_dims': (128, 256, 512, 1024),
			},
		'convnext_large': {
			'depths': (3, 3, 27, 3),
			'out_dims': (192, 384, 768, 1536),
			},
		'convnext_tiny_in22k': {
			'depths': (3, 3, 9, 3),
			'out_dims': (96, 192, 384, 768),
			},
		'convnext_small_in22k': {
			'depths': (3, 3, 27, 3),
			'out_dims': (96, 192, 384, 768),
			},
		'convnext_base_in22k': {
			'depths': (3, 3, 27, 3),
			'out_dims': (128, 256, 512, 1024),
			},
		'convnext_large_in22k': {
			'depths': (3, 3, 27, 3),
			'out_dims': (192, 384, 768, 1536),
			},
		'convnext_xlarge_in22k': {
			'depths': (3, 3, 27, 3),
			'out_dims': (256, 512, 1024, 2048),
			},
		'convnext_tiny_in22ft1k': {
			'depths': (3, 3, 9, 3),
			'out_dims': (96, 192, 384, 768),
			},
		'convnext_small_in22ft1k': {
			'depths': (3, 3, 27, 3),
			'out_dims': (96, 192, 384, 768),
			},
		'convnext_base_in22ft1k': {
			'depths': (3, 3, 27, 3),
			'out_dims': (128, 256, 512, 1024),
			},
		'convnext_large_in22ft1k': {
			'depths': (3, 3, 27, 3),
			'out_dims': (192, 384, 768, 1536),
			},
		'convnext_xlarge_in22ft1k': {
			'depths': (3, 3, 27, 3),
			'out_dims': (256, 512, 1024, 2048),
			},
		'convnext_tiny_384_in22ft1k': {
			'depths': (3, 3, 9, 3),
			'out_dims': (96, 192, 384, 768),
			},
		'convnext_small_384_in22ft1k': {
			'depths': (3, 3, 27, 3),
			'out_dims': (96, 192, 384, 768),
			},
		'convnext_base_384_in22ft1k': {
			'depths': (3, 3, 27, 3),
			'out_dims': (128, 256, 512, 1024),
			},
		'convnext_large_384_in22ft1k': {
			'depths': (3, 3, 27, 3),
			'out_dims': (192, 384, 768, 1536),
			},
		'convnext_xlarge_384_in22ft1k': {
			'depths': (3, 3, 27, 3),
			'out_dims': (256, 512, 1024, 2048),
			},
		}
	return ConvNeXt, configs
